# -*- coding: UTF-8 -*-
# pylint: disable=C0111,R0903

"""Displays currency exchange rates. Currently, displays currency between GBP and USD/EUR only.

Requires the following python packages:
    * requests

Parameters:
    * currency.interval: Interval in minutes between updates, default is 1.
    * currency.source: Source currency (ex. 'GBP', 'EUR'). Defaults to 'auto', which infers the local one from IP address.
    * currency.destination: Comma-separated list of destination currencies (defaults to 'USD,EUR')
    * currency.sourceformat: String format for source formatting; Defaults to '{}: {}' and has two variables,
      the base symbol and the rate list
    * currency.destinationdelimiter: Delimiter used for separating individual rates (defaults to '|')

Note: source and destination names right now must correspond to the names used by the API of https://markets.ft.com

contributed by `AntouanK <https://github.com/AntouanK>`_ - many thanks!
"""

import requests

try:
    from babel.numbers import format_currency
except ImportError:
    format_currency = None
import json
import os

import core.module
import core.widget
import core.decorators

import util.format
import util.location

SYMBOL = {"GBP": "£", "EUR": "€", "USD": "$", "JPY": "¥", "KRW": "₩"}
DEFAULT_DEST = "USD,EUR,auto"
DEFAULT_SRC = "GBP"

API_URL = "https://markets.ft.com/data/currencies/ajax/conversion?baseCurrency={}&comparison={}"


def load_country_to_currency():
    return [
        {"country": "Afghanistan", "currency_code": "AFN"},
        {"country": "Albania", "currency_code": "ALL"},
        {"country": "Algeria", "currency_code": "DZD"},
        {"country": "American Samoa", "currency_code": "USD"},
        {"country": "Andorra", "currency_code": "EUR"},
        {"country": "Angola", "currency_code": "AOA"},
        {"country": "Anguilla", "currency_code": "XCD"},
        {"country": "Antarctica", "currency_code": "XCD"},
        {"country": "Antigua and Barbuda", "currency_code": "XCD"},
        {"country": "Argentina", "currency_code": "ARS"},
        {"country": "Armenia", "currency_code": "AMD"},
        {"country": "Aruba", "currency_code": "AWG"},
        {"country": "Australia", "currency_code": "AUD"},
        {"country": "Austria", "currency_code": "EUR"},
        {"country": "Azerbaijan", "currency_code": "AZN"},
        {"country": "Bahamas", "currency_code": "BSD"},
        {"country": "Bahrain", "currency_code": "BHD"},
        {"country": "Bangladesh", "currency_code": "BDT"},
        {"country": "Barbados", "currency_code": "BBD"},
        {"country": "Belarus", "currency_code": "BYR"},
        {"country": "Belgium", "currency_code": "EUR"},
        {"country": "Belize", "currency_code": "BZD"},
        {"country": "Benin", "currency_code": "XOF"},
        {"country": "Bermuda", "currency_code": "BMD"},
        {"country": "Bhutan", "currency_code": "BTN"},
        {"country": "Bolivia", "currency_code": "BOB"},
        {"country": "Bosnia and Herzegovina", "currency_code": "BAM"},
        {"country": "Botswana", "currency_code": "BWP"},
        {"country": "Bouvet Island", "currency_code": "NOK"},
        {"country": "Brazil", "currency_code": "BRL"},
        {"country": "British Indian Ocean Territory", "currency_code": "USD"},
        {"country": "Brunei", "currency_code": "BND"},
        {"country": "Bulgaria", "currency_code": "BGN"},
        {"country": "Burkina Faso", "currency_code": "XOF"},
        {"country": "Burundi", "currency_code": "BIF"},
        {"country": "Cambodia", "currency_code": "KHR"},
        {"country": "Cameroon", "currency_code": "XAF"},
        {"country": "Canada", "currency_code": "CAD"},
        {"country": "Cape Verde", "currency_code": "CVE"},
        {"country": "Cayman Islands", "currency_code": "KYD"},
        {"country": "Central African Republic", "currency_code": "XAF"},
        {"country": "Chad", "currency_code": "XAF"},
        {"country": "Chile", "currency_code": "CLP"},
        {"country": "China", "currency_code": "CNY"},
        {"country": "Christmas Island", "currency_code": "AUD"},
        {"country": "Cocos (Keeling) Islands", "currency_code": "AUD"},
        {"country": "Colombia", "currency_code": "COP"},
        {"country": "Comoros", "currency_code": "KMF"},
        {"country": "Congo", "currency_code": "XAF"},
        {"country": "Cook Islands", "currency_code": "NZD"},
        {"country": "Costa Rica", "currency_code": "CRC"},
        {"country": "Croatia", "currency_code": "HRK"},
        {"country": "Cuba", "currency_code": "CUP"},
        {"country": "Cyprus", "currency_code": "EUR"},
        {"country": "Czech Republic", "currency_code": "CZK"},
        {"country": "Denmark", "currency_code": "DKK"},
        {"country": "Djibouti", "currency_code": "DJF"},
        {"country": "Dominica", "currency_code": "XCD"},
        {"country": "Dominican Republic", "currency_code": "DOP"},
        {"country": "East Timor", "currency_code": "USD"},
        {"country": "Ecuador", "currency_code": "ECS"},
        {"country": "Egypt", "currency_code": "EGP"},
        {"country": "El Salvador", "currency_code": "SVC"},
        {"country": "England", "currency_code": "GBP"},
        {"country": "Equatorial Guinea", "currency_code": "XAF"},
        {"country": "Eritrea", "currency_code": "ERN"},
        {"country": "Estonia", "currency_code": "EUR"},
        {"country": "Ethiopia", "currency_code": "ETB"},
        {"country": "Falkland Islands", "currency_code": "FKP"},
        {"country": "Faroe Islands", "currency_code": "DKK"},
        {"country": "Fiji Islands", "currency_code": "FJD"},
        {"country": "Finland", "currency_code": "EUR"},
        {"country": "France", "currency_code": "EUR"},
        {"country": "French Guiana", "currency_code": "EUR"},
        {"country": "French Polynesia", "currency_code": "XPF"},
        {"country": "French Southern territories", "currency_code": "EUR"},
        {"country": "Gabon", "currency_code": "XAF"},
        {"country": "Gambia", "currency_code": "GMD"},
        {"country": "Georgia", "currency_code": "GEL"},
        {"country": "Germany", "currency_code": "EUR"},
        {"country": "Ghana", "currency_code": "GHS"},
        {"country": "Gibraltar", "currency_code": "GIP"},
        {"country": "Greece", "currency_code": "EUR"},
        {"country": "Greenland", "currency_code": "DKK"},
        {"country": "Grenada", "currency_code": "XCD"},
        {"country": "Guadeloupe", "currency_code": "EUR"},
        {"country": "Guam", "currency_code": "USD"},
        {"country": "Guatemala", "currency_code": "QTQ"},
        {"country": "Guinea", "currency_code": "GNF"},
        {"country": "Guinea-Bissau", "currency_code": "CFA"},
        {"country": "Guyana", "currency_code": "GYD"},
        {"country": "Haiti", "currency_code": "HTG"},
        {"country": "Heard Island and McDonald Islands", "currency_code": "AUD"},
        {"country": "Holy See (Vatican City State)", "currency_code": "EUR"},
        {"country": "Honduras", "currency_code": "HNL"},
        {"country": "Hong Kong", "currency_code": "HKD"},
        {"country": "Hungary", "currency_code": "HUF"},
        {"country": "Iceland", "currency_code": "ISK"},
        {"country": "India", "currency_code": "INR"},
        {"country": "Indonesia", "currency_code": "IDR"},
        {"country": "Iran", "currency_code": "IRR"},
        {"country": "Iraq", "currency_code": "IQD"},
        {"country": "Ireland", "currency_code": "EUR"},
        {"country": "Israel", "currency_code": "ILS"},
        {"country": "Italy", "currency_code": "EUR"},
        {"country": "Ivory Coast", "currency_code": "XOF"},
        {"country": "Jamaica", "currency_code": "JMD"},
        {"country": "Japan", "currency_code": "JPY"},
        {"country": "Jordan", "currency_code": "JOD"},
        {"country": "Kazakhstan", "currency_code": "KZT"},
        {"country": "Kenya", "currency_code": "KES"},
        {"country": "Kiribati", "currency_code": "AUD"},
        {"country": "Kuwait", "currency_code": "KWD"},
        {"country": "Kyrgyzstan", "currency_code": "KGS"},
        {"country": "Laos", "currency_code": "LAK"},
        {"country": "Latvia", "currency_code": "LVL"},
        {"country": "Lebanon", "currency_code": "LBP"},
        {"country": "Lesotho", "currency_code": "LSL"},
        {"country": "Liberia", "currency_code": "LRD"},
        {"country": "Libyan Arab Jamahiriya", "currency_code": "LYD"},
        {"country": "Liechtenstein", "currency_code": "CHF"},
        {"country": "Lithuania", "currency_code": "LTL"},
        {"country": "Luxembourg", "currency_code": "EUR"},
        {"country": "Macao", "currency_code": "MOP"},
        {"country": "North Macedonia", "currency_code": "MKD"},
        {"country": "Madagascar", "currency_code": "MGF"},
        {"country": "Malawi", "currency_code": "MWK"},
        {"country": "Malaysia", "currency_code": "MYR"},
        {"country": "Maldives", "currency_code": "MVR"},
        {"country": "Mali", "currency_code": "XOF"},
        {"country": "Malta", "currency_code": "EUR"},
        {"country": "Marshall Islands", "currency_code": "USD"},
        {"country": "Martinique", "currency_code": "EUR"},
        {"country": "Mauritania", "currency_code": "MRO"},
        {"country": "Mauritius", "currency_code": "MUR"},
        {"country": "Mayotte", "currency_code": "EUR"},
        {"country": "Mexico", "currency_code": "MXN"},
        {"country": "Micronesia, Federated States of", "currency_code": "USD"},
        {"country": "Moldova", "currency_code": "MDL"},
        {"country": "Monaco", "currency_code": "EUR"},
        {"country": "Mongolia", "currency_code": "MNT"},
        {"country": "Montserrat", "currency_code": "XCD"},
        {"country": "Morocco", "currency_code": "MAD"},
        {"country": "Mozambique", "currency_code": "MZN"},
        {"country": "Myanmar", "currency_code": "MMR"},
        {"country": "Namibia", "currency_code": "NAD"},
        {"country": "Nauru", "currency_code": "AUD"},
        {"country": "Nepal", "currency_code": "NPR"},
        {"country": "Netherlands", "currency_code": "EUR"},
        {"country": "Netherlands Antilles", "currency_code": "ANG"},
        {"country": "New Caledonia", "currency_code": "XPF"},
        {"country": "New Zealand", "currency_code": "NZD"},
        {"country": "Nicaragua", "currency_code": "NIO"},
        {"country": "Niger", "currency_code": "XOF"},
        {"country": "Nigeria", "currency_code": "NGN"},
        {"country": "Niue", "currency_code": "NZD"},
        {"country": "Norfolk Island", "currency_code": "AUD"},
        {"country": "North Korea", "currency_code": "KPW"},
        {"country": "Northern Ireland", "currency_code": "GBP"},
        {"country": "Northern Mariana Islands", "currency_code": "USD"},
        {"country": "Norway", "currency_code": "NOK"},
        {"country": "Oman", "currency_code": "OMR"},
        {"country": "Pakistan", "currency_code": "PKR"},
        {"country": "Palau", "currency_code": "USD"},
        {"country": "Palestine", "currency_code": null},
        {"country": "Panama", "currency_code": "PAB"},
        {"country": "Papua New Guinea", "currency_code": "PGK"},
        {"country": "Paraguay", "currency_code": "PYG"},
        {"country": "Peru", "currency_code": "PEN"},
        {"country": "Philippines", "currency_code": "PHP"},
        {"country": "Pitcairn", "currency_code": "NZD"},
        {"country": "Poland", "currency_code": "PLN"},
        {"country": "Portugal", "currency_code": "EUR"},
        {"country": "Puerto Rico", "currency_code": "USD"},
        {"country": "Qatar", "currency_code": "QAR"},
        {"country": "Reunion", "currency_code": "EUR"},
        {"country": "Romania", "currency_code": "RON"},
        {"country": "Russian Federation", "currency_code": "RUB"},
        {"country": "Rwanda", "currency_code": "RWF"},
        {"country": "Saint Helena", "currency_code": "SHP"},
        {"country": "Saint Kitts and Nevis", "currency_code": "XCD"},
        {"country": "Saint Lucia", "currency_code": "XCD"},
        {"country": "Saint Pierre and Miquelon", "currency_code": "EUR"},
        {"country": "Saint Vincent and the Grenadines", "currency_code": "XCD"},
        {"country": "Samoa", "currency_code": "WST"},
        {"country": "San Marino", "currency_code": "EUR"},
        {"country": "Sao Tome and Principe", "currency_code": "STD"},
        {"country": "Saudi Arabia", "currency_code": "SAR"},
        {"country": "Scotland", "currency_code": "GBP"},
        {"country": "Senegal", "currency_code": "XOF"},
        {"country": "Seychelles", "currency_code": "SCR"},
        {"country": "Sierra Leone", "currency_code": "SLL"},
        {"country": "Singapore", "currency_code": "SGD"},
        {"country": "Slovakia", "currency_code": "EUR"},
        {"country": "Slovenia", "currency_code": "EUR"},
        {"country": "Solomon Islands", "currency_code": "SBD"},
        {"country": "Somalia", "currency_code": "SOS"},
        {"country": "South Africa", "currency_code": "ZAR"},
        {
            "country": "South Georgia and the South Sandwich Islands",
            "currency_code": "GBP",
        },
        {"country": "South Korea", "currency_code": "KRW"},
        {"country": "South Sudan", "currency_code": "SSP"},
        {"country": "Spain", "currency_code": "EUR"},
        {"country": "SriLanka", "currency_code": "LKR"},
        {"country": "Sudan", "currency_code": "SDG"},
        {"country": "Suriname", "currency_code": "SRD"},
        {"country": "Svalbard and Jan Mayen", "currency_code": "NOK"},
        {"country": "Swaziland", "currency_code": "SZL"},
        {"country": "Sweden", "currency_code": "SEK"},
        {"country": "Switzerland", "currency_code": "CHF"},
        {"country": "Syria", "currency_code": "SYP"},
        {"country": "Tajikistan", "currency_code": "TJS"},
        {"country": "Tanzania", "currency_code": "TZS"},
        {"country": "Thailand", "currency_code": "THB"},
        {"country": "The Democratic Republic of Congo", "currency_code": "CDF"},
        {"country": "Togo", "currency_code": "XOF"},
        {"country": "Tokelau", "currency_code": "NZD"},
        {"country": "Tonga", "currency_code": "TOP"},
        {"country": "Trinidad and Tobago", "currency_code": "TTD"},
        {"country": "Tunisia", "currency_code": "TND"},
        {"country": "Turkey", "currency_code": "TRY"},
        {"country": "Turkmenistan", "currency_code": "TMT"},
        {"country": "Turks and Caicos Islands", "currency_code": "USD"},
        {"country": "Tuvalu", "currency_code": "AUD"},
        {"country": "Uganda", "currency_code": "UGX"},
        {"country": "Ukraine", "currency_code": "UAH"},
        {"country": "United Arab Emirates", "currency_code": "AED"},
        {"country": "United Kingdom", "currency_code": "GBP"},
        {"country": "United States", "currency_code": "USD"},
        {"country": "United States Minor Outlying Islands", "currency_code": "USD"},
        {"country": "Uruguay", "currency_code": "UYU"},
        {"country": "Uzbekistan", "currency_code": "UZS"},
        {"country": "Vanuatu", "currency_code": "VUV"},
        {"country": "Venezuela", "currency_code": "VEF"},
        {"country": "Vietnam", "currency_code": "VND"},
        {"country": "Virgin Islands, British", "currency_code": "USD"},
        {"country": "Virgin Islands, U.S.", "currency_code": "USD"},
        {"country": "Wales", "currency_code": "GBP"},
        {"country": "Wallis and Futuna", "currency_code": "XPF"},
        {"country": "Western Sahara", "currency_code": "MAD"},
        {"country": "Yemen", "currency_code": "YER"},
        {"country": "Yugoslavia", "currency_code": null},
        {"country": "Zambia", "currency_code": "ZMW"},
        {"country": "Zimbabwe", "currency_code": "ZWD"},
    ]


class Module(core.module.Module):
    @core.decorators.every(minutes=5)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.price))

        self.__data = []

        src = self.parameter("source", DEFAULT_SRC)
        if src == "auto":
            self.__base = self.find_local_currency()
        else:
            self.__base = src

        self.__symbols = []
        for d in util.format.aslist(self.parameter("destination", DEFAULT_DEST)):
            if d == "auto":
                new = self.find_local_currency()
            else:
                new = d
            if new != self.__base:
                self.__symbols.append(new)

    def price(self, widget):
        if len(self.__data) == 0:
            return "?"

        rates = []
        for sym, rate in self.__data:
            rate_float = float(rate.replace(",", ""))
            if format_currency:
                rates.append(format_currency(rate_float, sym))
            else:
                rate = self.fmt_rate(rate)
                rates.append("{}{}".format(rate, SYMBOL[sym] if sym in SYMBOL else sym))

        basefmt = "{}".format(self.parameter("sourceformat", "{}={}"))
        ratefmt = "{}".format(self.parameter("destinationdelimiter", "="))

        if format_currency:
            base_val = format_currency(1, self.__base)
        else:
            base_val = "1{}".format(
                SYMBOL[self.__base] if self.__base in SYMBOL else self.__base
            )

        return basefmt.format(base_val, ratefmt.join(rates))

    def update(self):
        self.__data = []
        for symbol in self.__symbols:
            url = API_URL.format(self.__base, symbol)
            try:
                response = requests.get(url).json()
                self.__data.append((symbol, response["data"]["exchangeRate"]))
            except Exception:
                pass

    def find_local_currency(self):
        """Use geolocation lookup to find local currency"""
        try:
            country = util.location.country()
            currency_map = load_country_to_currency()
            return currency_map.get(country, DEFAULT_SRC)
        except:
            return DEFAULT_SRC

    def fmt_rate(self, rate):
        float_rate = float(rate.replace(",", ""))
        if not 0.01 < float_rate < 100:
            ret = rate
        else:
            ret = "%.3g" % float_rate

        return ret


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays sunrise and sunset times

Requires the following python packages:
    * requests
    * suntime

Parameters:
    * cpu.lat : Latitude of your location
    * cpu.lon : Longitude of your location

(if none of those are set, location is determined automatically via location APIs)

contributed by `lonesomebyte537 <https://github.com/lonesomebyte537>`_ - many thanks!
"""

from suntime import Sun, SunTimeException
import requests
from dateutil.tz import tzlocal

import datetime

import core.module
import core.widget
import core.decorators

import util.location


class Module(core.module.Module):
    @core.decorators.every(hours=1)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.suntimes))

        lat = self.parameter("lat", None)
        lon = self.parameter("lon", None)
        self.__sun = None

        if not lat or not lon:
            lat, lon = util.location.coordinates()
        if lat and lon:
            self.__sun = Sun(float(lat), float(lon))

    def suntimes(self, _):
        if self.__sunset and self.__sunrise:
            if self.__isup:
                return "\u21A7{} \u21A5{}".format(
                    self.__sunset.strftime("%H:%M"), self.__sunrise.strftime("%H:%M")
                )
            return "\u21A5{} \u21A7{}".format(
                self.__sunrise.strftime("%H:%M"), self.__sunset.strftime("%H:%M")
            )
        return "n/a"

    def __calculate_times(self):
        self.__isup = False

        order_matters = True

        try:
            self.__sunrise = self.__sun.get_local_sunrise_time()
        except SunTimeException:
            self.__sunrise = "no sunrise"
            order_matters = False

        try:
            self.__sunset = self.__sun.get_local_sunset_time()
        except SunTimeException:
            self.__sunset = "no sunset"
            order_matters = False

        if not order_matters:
            return

        now = datetime.datetime.now(tz=tzlocal())
        if now > self.__sunset:
            tomorrow = (now + datetime.timedelta(days=1)).date()
            try:
                self.__sunrise = self.__sun.get_local_sunrise_time(tomorrow)
                self.__sunset = self.__sun.get_local_sunset_time(tomorrow)
            except SunTimeException:
                self.__sunrise = "no sunrise"
                self.__sunset = "no sunset"

        elif now > self.__sunrise:
            tomorrow = (now + datetime.timedelta(days=1)).date()
            try:
                self.__sunrise = self.__sun.get_local_sunrise_time(tomorrow)
            except SunTimeException:
                self.__sunrise = "no sunrise"
                return
            self.__isup = True

    def update(self):
        self.__calculate_times()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# -*- coding: utf-8 -*-

"""Displays GPU name, temperature and memory usage.

Parameters:
   * nvidiagpu.format: Format string (defaults to '{name}: {temp}°C %{usedmem}/{totalmem} MiB')
     Available values are: {name} {temp} {mem_used} {mem_total} {fanspeed} {clock_gpu} {clock_mem}

Requires nvidia-smi

contributed by `RileyRedpath <https://github.com/RileyRedpath>`_ - many thanks!
"""

import core.module
import core.widget

import util.cli
import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.utilization))

        self.__utilization = "Not found: 0 0/0"

    def utilization(self, widget):
        return self.__utilization

    def hidden(self):
        return "not found" in self.__utilization

    def update(self):
        sp = util.cli.execute("nvidia-smi -q", ignore_errors=True)

        title = ""
        usedMem = ""
        totalMem = ""
        temp = ""
        name = "not found"
        clockMem = ""
        clockGpu = ""
        fanspeed = ""
        for item in sp.split("\n"):
            try:
                key, val = item.split(":")
                key, val = key.strip(), val.strip()
                if title == "Clocks":
                    if key == "Graphics":
                        clockGpu = val.split(" ")[0]
                    elif key == "Memory":
                        clockMem = val.split(" ")[0]
                if title == "FB Memory Usage":
                    if key == "Total":
                        totalMem = val.split(" ")[0]
                    elif key == "Used":
                        usedMem = val.split(" ")[0]
                elif key == "GPU Current Temp":
                    temp = val.split(" ")[0]
                elif key == "Product Name":
                    name = val
                elif key == "Fan Speed":
                    fanspeed = val.split(" ")[0]

            except:
                title = item.strip()

        str_format = self.parameter(
            "format", "{name}: {temp}°C {mem_used}/{mem_total} MiB"
        )
        self.__utilization = str_format.format(
            name=name,
            temp=temp,
            mem_used=usedMem,
            mem_total=totalMem,
            clock_gpu=clockGpu,
            clock_mem=clockMem,
            fanspeed=fanspeed,
        )


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the system uptime.

contributed by `ccoors <https://github.com/ccoors>`_ - many thanks!
"""

from datetime import timedelta

import core.module
import core.widget


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.output))
        self.__uptime = ""

    def output(self, _):
        return "{}".format(self.__uptime)

    def update(self):
        with open("/proc/uptime", "r") as f:
            uptime_seconds = int(float(f.readline().split()[0]))
            self.__uptime = timedelta(seconds=uptime_seconds)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

""" Displays the VPN profile that is currently in use.

    Left click opens a popup menu that lists all available VPN profiles and allows to establish
    a VPN connection using that profile.

    Prerequisites:
         * tk python library (usually python-tk or python3-tk, depending on your distribution)
         * nmcli needs to be installed and configured properly.
           To quickly test, whether nmcli is working correctly, type 'nmcli -g NAME,TYPE,DEVICE con' which
           lists all the connection profiles that are configured. Make sure that your VPN profile is in that list!

           e.g: to import a openvpn profile via nmcli:
           `sudo nmcli connection import type openvpn file </path/to/your/openvpn/profile.ovpn>`

contributed by `bbernhard <https://github.com/bbernhard>`_ - many thanks!
"""

import logging
import functools

import core.module
import core.widget
import core.input

import util.cli
import util.popup


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.vpn_status))

        self.__connected_vpn_profile = None
        self.__selected_vpn_profile = None

        res = util.cli.execute("nmcli -g NAME,TYPE c")
        lines = res.splitlines()

        self.__vpn_profiles = []
        for line in lines:
            info = line.split(":")
            try:
                if self.__isvpn(info[1]):
                    self.__vpn_profiles.append(info[0])
            except:
                pass

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.popup)

    def __isvpn(self, connection_type):
        return connection_type in ["vpn", "wireguard"]

    def update(self):
        try:
            res = util.cli.execute("nmcli -g NAME,TYPE,DEVICE con")
            lines = res.splitlines()
            self.__connected_vpn_profile = None
            for line in lines:
                info = line.split(":")
                if self.__isvpn(info[1]) and info[2] != "":
                    self.__connected_vpn_profile = info[0]

        except Exception as e:
            logging.exception("Could not get VPN status")
            self.__connected_vpn_profile = None

    def vpn_status(self, widget):
        if self.__connected_vpn_profile is None:
            return "off"
        return self.__connected_vpn_profile

    def __on_vpndisconnect(self):
        try:
            util.cli.execute(
                "nmcli c down '{vpn}'".format(vpn=self.__connected_vpn_profile)
            )
            self.__connected_vpn_profile = None
        except Exception as e:
            logging.exception("Could not disconnect VPN connection")

    def __on_vpnconnect(self, name):
        self.__selected_vpn_profile = name

        try:
            util.cli.execute(
                "nmcli c up '{vpn}'".format(vpn=self.__selected_vpn_profile)
            )
            self.__connected_vpn_profile = name
        except Exception as e:
            logging.exception("Could not establish VPN connection")
            self.__connected_vpn_profile = None

    def popup(self, widget):
        menu = util.popup.menu()

        if self.__connected_vpn_profile is not None:
            menu.add_menuitem("Disconnect", callback=self.__on_vpndisconnect)
        for vpn_profile in self.__vpn_profiles:
            if (
                self.__connected_vpn_profile is not None
                and self.__connected_vpn_profile == vpn_profile
            ):
                continue
            menu.add_menuitem(
                vpn_profile,
                callback=functools.partial(self.__on_vpnconnect, vpn_profile),
            )
        menu.show(widget)

    def state(self, widget):
        return []


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Shows a widget for each connected screen and allows the user to loop through different orientations.

Requires the following executable:
    * xrandr
"""

import core.module
import core.input

import util.cli

possible_orientations = ["normal", "left", "inverted", "right"]


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

    def update(self):
        widgets = self.widgets()
        for line in util.cli.execute("xrandr -q").split("\n"):
            if not " connected" in line:
                continue
            display = line.split(" ", 2)[0]

            orientation = "normal"
            for curr_orient in possible_orientations:
                if (line.split(" ")).count(curr_orient) > 1:
                    orientation = curr_orient
                    break

            widget = self.widget(display)
            if not widget:
                widget = self.add_widget(full_text=display, name=display)
                core.input.register(
                    widget, button=core.input.LEFT_MOUSE, cmd=self.__toggle
                )
            widget.set("orientation", orientation)
            widgets.append(widget)

    def state(self, widget):
        return widget.get("orientation", "normal")

    def __toggle(self, event):
        widget = self.widget_by_id(event["instance"])

        # compute new orientation based on current orientation
        idx = possible_orientations.index(widget.get("orientation"))
        idx = (idx + 1) % len(possible_orientations)
        new_orientation = possible_orientations[idx]

        widget.set("orientation", new_orientation)

        util.cli.execute(
            "xrandr --output {} --rotation {}".format(widget.name, new_orientation)
        )


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

import core.widget
import core.module
import core.config


class SampleModule(core.module.Module):
    def __init__(self, widgets, config=core.config.Config([]), theme=None):
        super().__init__(config, theme, widgets)
        self.states = []

    def update(self):
        if self.fail:
            raise Exception(self.error)
        pass

    def state(self, widget):
        return self.states


@pytest.fixture
def widget_a():
    return core.widget.Widget("some random value")


# class widget(unittest.TestCase):
#    def setUp(self):
#        self.someValue = "some random value"
#        self.someOtherValue = "some different value"
#        self.callbackReturnValue = "callback return value"
#        self.someWidget = core.widget.Widget(full_text=self.someValue)
#        self.someCallback = unittest.mock.MagicMock(
#            return_value=self.callbackReturnValue
#        )
#
#        self.assertNotEqual(self.someValue, self.someOtherValue)


def test_text_fulltext():
    widget = core.widget.Widget(full_text="this is some value")
    assert widget.full_text() == "this is some value"


def test_set_fulltext(widget_a):
    assert widget_a.full_text() != "new value"
    widget_a.full_text("new value")
    assert widget_a.full_text() == "new value"


def test_callable_fulltext(mocker):
    callback = mocker.MagicMock(return_value="callback returns")
    widget = core.widget.Widget(full_text=callback)
    assert widget.full_text() == "callback returns"
    callback.assert_called_once_with(widget)


def test_set_callable_fulltext(mocker, widget_a):
    callback = mocker.MagicMock(return_value="this is a test")
    widget_a.full_text(callback)
    assert widget_a.full_text() == "this is a test"
    callback.assert_called_once_with(widget_a)


def test_state_defaults_to_empty(widget_a):
    assert widget_a.state() == []


def test_single_widget_state(widget_a):
    widget_a.set("state", "state1")
    assert widget_a.state() == ["state1"]


def test_multiple_widget_states(widget_a):
    widget_a.set("state", ["state1", "state2"])
    assert widget_a.state() == ["state1", "state2"]


def test_widget_module_state(widget_a):
    module = SampleModule(widgets=widget_a)
    widget_a.set("state", ["state1", "state2"])

    module.states = "x"
    assert widget_a.state() == ["state1", "state2", "x"]

    module.states = ["a", "b"]
    assert widget_a.state() == ["state1", "state2", "a", "b"]


def test_multiple_widget_themes():
    widget1 = core.widget.Widget(full_text="a")
    widget2 = core.widget.Widget(full_text="b")
    widget3 = core.widget.Widget(full_text="c")

    module = SampleModule(widgets=[widget1, widget2, widget3])
    module.set("theme.test", "1,2,3")
    module.set("theme.test2", "x")

    assert widget1.theme("test") == "1"
    assert widget2.theme("test") == "2"
    assert widget3.theme("test") == "3"

    assert widget1.theme("test2") == "x"
    assert widget2.theme("test2") == None
    assert widget3.theme("test2") == None


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Toggle twmn notifications.

contributed by `Pseudonick47 <https://github.com/Pseudonick47>`_ - many thanks!
"""

import core.module
import core.widget
import core.input
import core.decorators

import util.cli


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(""))

        self.__paused = False
        # Make sure that twmn is currently not paused
        util.cli.execute("killall -SIGUSR2 twmnd", ignore_errors=True)
        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.toggle_status)

    def toggle_status(self, event):
        self.__paused = not self.__paused

        try:
            if self.__paused:
                util.cli.execute("systemctl --user start twmnd")
            else:
                util.cli.execute("systemctl --user stop twmnd")
        except:
            self.__paused = not self.__paused  # toggling failed

    def state(self, widget):
        if self.__paused:
            return ["muted"]
        return ["unmuted"]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Display HTTP status code

Parameters:
    * http__status.label: Prefix label (optional)
    * http__status.target: Target to retrieve the HTTP status from
    * http__status.expect: Expected HTTP status

contributed by `valkheim <https://github.com/valkheim>`_ - many thanks!
"""

from requests import head

import psutil

import core.module
import core.widget
import core.decorators


class Module(core.module.Module):
    UNK = "UNK"

    @core.decorators.every(seconds=30)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.output))

        self.__label = self.parameter("label")
        self.__target = self.parameter("target")
        self.__expect = self.parameter("expect", "200")

    def labelize(self, s):
        if self.__label is None:
            return s
        return "{}: {}".format(self.__label, s)

    def getStatus(self):
        try:
            res = head(self.__target)
        except Exception as e:
            print(e)
            return self.UNK
        else:
            status = str(res.status_code)
            return status

    def getOutput(self):
        if self.__status == self.__expect:
            return self.labelize(self.__status)
        else:
            reason = " != {}".format(self.__expect)
            return self.labelize("{}{}".format(self.__status, reason))

    def output(self, widget):
        return self.__output

    def update(self):
        self.__status = self.getStatus()
        self.__output = self.getOutput()

    def state(self, widget):
        if self.__status == self.UNK:
            return "warning"
        if self.__status != self.__expect:
            return "critical"
        return self.__output


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the result of a notmuch count query
   default : unread emails which path do not contained 'Trash' (notmuch count 'tag:unread AND NOT path:/.*Trash.*/')

Parameters:
    * notmuch_count.query: notmuch count query to show result 

Errors:
    if the notmuch query failed, the shown value is  -1

Dependencies:
    notmuch (https://notmuchmail.org/)

contributed by `abdoulayeYATERA <https://github.com/abdoulayeYATERA>`_ - many thanks!
"""

import os

import core.module
import core.widget

import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.output))

        self.__notmuch_count_query = self.parameter(
            "query", "tag:unread AND NOT path:/.*Trash.*/"
        )

    def output(self, widget):
        return self.__notmuch_count

    def state(self, widgets):
        if self.__notmuch_count == 0:
            return "empty"
        return "items"

    def update(self):
        try:
            self.__notmuch_count = util.cli.execute(
                "notmuch count {}".format(self.__notmuch_count_query)
            ).strip()
        except Exception:
            self.__notmuch_count = "n/a"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Displays count of running libvirt VMs.

Required the following python packages:
        * libvirt

contributed by `maxpivo <https://github.com/maxpivo>`_ - many thanks!
"""

import sys
import libvirt

import core.module
import core.widget
import core.input
import core.decorators


class Module(core.module.Module):
    @core.decorators.every(seconds=10)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.status))

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd="virt-manager")

    def status(self, _):
        conn = libvirt.openReadOnly(None)
        if conn == None:
            return "Failed to open connection to the hypervisor"
        return "VMs %s" % (conn.numOfDomains())


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Displays public IP address
"""

import core.module
import core.widget
import core.decorators

import util.location


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.public_ip))

        self.__ip = ""

    def public_ip(self, widget):
        return self.__ip

    def update(self):
        try:
            self.__ip = util.location.public_ip()
        except Exception:
            self.__ip = "n/a"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import copy


def merge(target, *args):
    """Merges arbitrary data - copied from http://blog.impressiver.com/post/31434674390/deep-merge-multiple-python-dicts

    :param target: the data structure to fill
    :param args: a list of data structures to merge into target

    :return: target, with all data in args merged into it
    :rtype: whatever type was originally passed in
    """
    if len(args) > 1:
        for item in args:
            merge(target, item)
        return target

    item = args[0]
    if not isinstance(item, dict):
        return item
    for key, value in item.items():
        if key in target and isinstance(target[key], dict):
            merge(target[key], value)
        else:
            if not key in target:
                target[key] = copy.deepcopy(value)
    return target


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Displays the number of pending tasks in TaskWarrior.

Requires the following library:
    * taskw

Parameters:
    * taskwarrior.taskrc : path to the taskrc file (defaults to ~/.taskrc)


contributed by `chdorb <https://github.com/chdorb>`_ - many thanks!
"""

from taskw import TaskWarrior

import core.module
import core.widget
import core.decorators


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.output))

        self.__pending_tasks = "0"

    def update(self):
        """Return a string with the number of pending tasks from TaskWarrior."""
        try:
            taskrc = self.parameter("taskrc", "~/.taskrc")
            w = TaskWarrior(config_filename=taskrc)
            pending_tasks = w.filter_tasks({"status": "pending"})
            self.__pending_tasks = str(len(pending_tasks))
        except:
            self.__pending_tasks = "n/a"

    def output(self, _):
        """Format the task counter to output in bumblebee."""
        return "{}".format(self.__pending_tasks)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Displays bluetooth status. Left mouse click launches manager app,
right click toggles bluetooth. Needs dbus-send to toggle bluetooth state and
python-dbus to count the number of connections

Parameters:
    * bluetooth.manager : application to launch on click (blueman-manager)

contributed by `martindoublem <https://github.com/martindoublem>`_ - many thanks!
"""


import os
import re
import subprocess
import dbus
import dbus.mainloop.glib
import logging

import core.module
import core.widget
import core.input

import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.status))

        self.manager = self.parameter("manager", "blueman-manager")
        self._status = "Off"
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        self._bus = dbus.SystemBus()

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.manager)
        core.input.register(self, button=core.input.RIGHT_MOUSE, cmd=self._toggle)

    def status(self, widget):
        """Get status."""
        return self._status

    def update(self):
        """Update current state."""
        state = len(
            subprocess.run(["bluetoothctl", "list"], stdout=subprocess.PIPE).stdout
        )
        if state > 0:
            connected_devices = self.get_connected_devices()
            self._status = "On - {}".format(connected_devices)
        else:
            self._status = "Off"
            adapters_cmd = "rfkill list | grep Bluetooth"
            if not len(
                subprocess.run(adapters_cmd, shell=True, stdout=subprocess.PIPE).stdout
            ):
                self._status = "No Adapter Found"
        return

    def _toggle(self, widget=None):
        """Toggle bluetooth state."""
        if "On" in self._status:
            state = "false"
        else:
            state = "true"

        cmd = (
            "dbus-send --system --print-reply --dest=org.blueman.Mechanism /org/blueman/mechanism org.blueman.Mechanism.SetRfkillState boolean:%s"
            % state
        )

        logging.debug("bt: toggling bluetooth")
        core.util.execute(cmd)

    def state(self, widget):
        """Get current state."""
        state = []

        if self._status == "No Adapter Found":
            state.append("critical")
        elif self._status == "On - 0":
            state.append("warning")
        elif "On" in self._status and not (self._status == "On - 0"):
            state.append("ON")
        else:
            state.append("critical")
        return state

    def get_connected_devices(self):
        devices = 0
        objects = dbus.Interface(
            self._bus.get_object("org.bluez", "/"), "org.freedesktop.DBus.ObjectManager"
        ).GetManagedObjects()
        for path, interfaces in objects.items():
            if "org.bluez.Device1" in interfaces:
                if dbus.Interface(
                    self._bus.get_object("org.bluez", path),
                    "org.freedesktop.DBus.Properties",
                ).Get("org.bluez.Device1", "Connected"):
                    devices += 1
        return devices


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# -*- coding: UTF-8 -*-

"""Displays sensor temperature and CPU frequency

Parameters:

    * sensors2.chip: 'sensors -u' compatible filter for chip to display (default to empty - show all chips)
    * sensors2.showcpu: Enable or disable CPU frequency display (default: true)
    * sensors2.showtemp: Enable or disable temperature display (default: true)
    * sensors2.showfan: Enable or disable fan display (default: true)
    * sensors2.showother: Enable or display 'other' sensor readings (default: false)
    * sensors2.showname: Enable or disable show of sensor name (default: false)
    * sensors2.chip_include: Comma-separated list of chip to include (defaults to '' will include all by default, example: 'coretemp,bat')
    * sensors2.chip_exclude:Comma separated list of chip to exclude (defaults to '' will exlude none by default)
    * sensors2.field_include: Comma separated list of chip to include (defaults to '' will include all by default, example: 'temp,fan')
    * sensors2.field_exclude: Comma separated list of chip to exclude (defaults to '' will exclude none by default)
    * sensors2.chip_field_exclude: Comma separated list of chip field to exclude (defaults to '' will exclude none by default, example: 'coretemp-isa-0000.temp1,coretemp-isa-0000.fan1')
    * sensors2.chip_field_include: Comma-separated list of adaper field to include (defaults to '' will include all by default)
"""

import re

import core.module
import core.widget

import util.cli
import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self.__chip = self.parameter("chip", "")
        self.__data = {}
        self.__update()

        self.__create_widgets()

    def update(self):
        self.__update()
        for widget in self.widgets():
            self.__update_widget(widget)

    def state(self, widget):
        widget_type = widget.get("type", "")
        try:
            data = self.__data[widget.get("adapter")][widget.get("package")][
                widget.get("field")
            ]
            if "crit" in data and float(data["input"]) > float(data["crit"]):
                return ["critical", widget_type]
            if "max" in data and float(data["input"]) > float(data["max"]):
                return ["warning", widget_type]
        except:
            pass
        return [widget_type]

    def __create_widgets(self):
        show_temp = util.format.asbool(self.parameter("showtemp", True))
        show_fan = util.format.asbool(self.parameter("showfan", True))
        show_other = util.format.asbool(self.parameter("showother", False))
        include_chip = tuple(
            filter(len, util.format.aslist(self.parameter("chip_include", "")))
        )
        exclude_chip = tuple(
            filter(len, util.format.aslist(self.parameter("chip_exclude", "")))
        )
        include_field = tuple(
            filter(len, util.format.aslist(self.parameter("field_include", "")))
        )
        exclude_field = tuple(
            filter(len, util.format.aslist(self.parameter("field_exclude", "")))
        )
        include_chip_field = tuple(
            filter(len, util.format.aslist(self.parameter("chip_field_include", "")))
        )
        exclude_chip_field = tuple(
            filter(len, util.format.aslist(self.parameter("chip_field_exclude", "")))
        )

        if util.format.asbool(self.parameter("showcpu", True)):
            widget = self.add_widget(full_text=self.__cpu)
            widget.set("type", "cpu")

        for adapter in self.__data:
            if include_chip or exclude_chip:
                if include_chip:
                    if all([chip not in adapter for chip in include_chip]):
                        continue
                else:
                    if any([chip in adapter for chip in exclude_chip]):
                        continue

            if include_chip_field:
                try:
                    if all(
                        [i.split(".")[0] not in adapter for i in include_chip_field]
                    ):
                        continue
                except:
                    pass

            for package in self.__data[adapter]:
                if util.format.asbool(self.parameter("showname", False)):
                    widget = self.add_widget(full_text=package)
                    widget.set("data", self.__data[adapter][package])
                    widget.set("package", package)
                    widget.set("field", "")
                    widget.set("adapter", adapter)
                for field in self.__data[adapter][package]:

                    if include_field or exclude_field:
                        if include_field:
                            if all(
                                [included not in field for included in include_field]
                            ):
                                continue
                        else:
                            if any([excluded in field for excluded in exclude_field]):
                                continue

                    try:
                        if include_chip_field or exclude_chip_field:
                            if include_chip_field:
                                if all(
                                    [
                                        i.split(".")[1] not in field
                                        for i in include_chip_field
                                        if i.split(".")[0] in adapter
                                    ]
                                ):
                                    continue
                            else:
                                if any(
                                    [
                                        i.split(".")[1] in field
                                        for i in exclude_chip_field
                                        if i.split(".")[0] in adapter
                                    ]
                                ):
                                    continue
                    except:
                        pass

                    widget = None
                    if "temp" in field and show_temp:
                        # seems to be a temperature
                        widget = self.add_widget()
                        widget.set("type", "temp")
                    if "fan" in field and show_fan:
                        # seems to be a fan
                        widget = self.add_widget()
                        widget.set("type", "fan")
                    elif show_other:
                        # everything else
                        widget = self.add_widget()
                        widget.set("type", "other")
                    if widget:
                        widget.set("package", package)
                        widget.set("field", field)
                        widget.set("adapter", adapter)

    def __update_widget(self, widget):
        if widget.get("field", "") == "":
            return  # nothing to do
        data = self.__data[widget.get("adapter")][widget.get("package")][
            widget.get("field")
        ]
        if "temp" in widget.get("field"):
            widget.full_text("{:0.01f}°C".format(data["input"]))
        elif "fan" in widget.get("field"):
            widget.full_text("{:0.0f}RPM".format(data["input"]))
        else:
            widget.full_text("{:0.0f}".format(data["input"]))

    def __update(self):
        output = util.cli.execute(
            "sensors -u {}".format(self.__chip), ignore_errors=True
        )
        self.__data = self.__parse(output)

    def __parse(self, data):
        output = {}
        package = ""
        adapter = None
        chip = None
        for line in data.split("\n"):
            if "Adapter" in line:
                # new adapter
                line = line.replace("Adapter: ", "")
                output[chip + " " + line] = {}
                adapter = chip + " " + line
            chip = line  # default - line before adapter is always the chip
            if not adapter:
                continue
            key, value = (line.split(":") + ["", ""])[:2]
            if not line.startswith(" "):
                # assume this starts a new package
                if package in output[adapter] and output[adapter][package] == {}:
                    del output[adapter][package]
                output[adapter][key] = {}
                package = key
            else:
                # feature for this chip
                try:
                    name, variant = (key.strip().split("_", 1) + ["", ""])[:2]
                    if not name in output[adapter][package]:
                        output[adapter][package][name] = {}
                    if variant:
                        output[adapter][package][name][variant] = {}
                    output[adapter][package][name][variant] = float(value)
                except Exception as e:
                    pass
        return output

    def __cpu(self, _):
        mhz = None
        try:
            output = open(
                "/sys/devices/system/cpu/cpufreq/policy0/scaling_cur_freq"
            ).read()
            mhz = int(float(output) / 1000.0)
        except:
            output = open("/proc/cpuinfo").read()
            m = re.search(r"cpu MHz\s+:\s+(\d+)", output)
            if m:
                mhz = int(m.group(1))
            else:
                m = re.search(r"BogoMIPS\s+:\s+(\d+)", output)
                if m:
                    return "{} BogoMIPS".format(int(m.group(1)))
        if not mhz:
            return "n/a"

        if mhz < 1000:
            return "{} MHz".format(mhz)
        else:
            return "{:0.01f} GHz".format(float(mhz) / 1000.0)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Shows that debug is enabled"""

import platform

import core.module
import core.widget
import core.decorators


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.full_text))

    def full_text(self, widgets):
        return "debug"

    def state(self, widget):
        return "warning"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import os
import importlib
import logging
import threading

import core.config
import core.input
import core.widget
import core.decorators

import util.format

try:
    error = ModuleNotFoundError("")
except Exception as e:
    ModuleNotFoundError = Exception

log = logging.getLogger(__name__)


"""Loads a module by name

:param module_name: Name of the module to load
:param config: Configuration to apply to the module (defaults to an empty configuration)
:param theme: Theme for this module, defaults to None, which means whatever is configured in "config"

:return: A module object representing the module, or an Error module if loading failed
:rtype: class bumblebee_status.module.Module
"""


def load(module_name, config=core.config.Config([]), theme=None):
    error = None
    module_short, alias = (module_name.split(":") + [module_name])[0:2]
    config.set("__alias__", alias)
    for namespace in ["core", "contrib"]:
        try:
            mod = importlib.import_module(
                "modules.{}.{}".format(namespace, module_short)
            )
            log.debug(
                "importing {} from {}.{}".format(module_short, namespace, module_short)
            )
            return getattr(mod, "Module")(config, theme)
        except ImportError as e:
            log.debug("failed to import {}: {}".format(module_name, e))
            error = e
    log.fatal("failed to import {}: {}".format(module_name, error))
    return Error(config=config, module=module_name, error=error)


class Module(core.input.Object):
    """Represents a module (single piece of functionality) of the bar

    :param config: Configuration to apply to the module (defaults to an empty configuration)
    :param theme: Theme for this module, defaults to None, which means whatever is configured in "config"
    :param widgets: A list of bumblebee_status.widget.Widget objects that the module is comprised of
    """

    def __init__(self, config=core.config.Config([]), theme=None, widgets=[]):
        super().__init__()
        self.background = False
        self.__thread = None
        self.__config = config
        self.__widgets = widgets if isinstance(widgets, list) else [widgets]

        self.module_name = self.__module__.split(".")[-1]
        self.name = self.module_name
        self.alias = self.__config.get("__alias__", None)
        self.id = self.alias if self.alias else self.name
        self.next_update = None

        self.theme = theme

        for widget in self.__widgets:
            widget.module = self

    """Override this to determine when to show this module

    :return: True if the module should be hidden, False otherwise
    :rtype: boolean
    """

    def hidden(self):
        return False

    """Retrieve CLI/configuration parameters for this module. For example, if
    the module is called "test" and the user specifies "-p test.x=123" on the
    commandline, using self.parameter("x") retrieves the value 123.

    :param key: Name of the parameter to retrieve
    :param default: Default value, if parameter is not set by user (defaults to None)

    :return: Parameter value, or default
    :rtype: string
    """

    def parameter(self, key, default=None):
        value = default

        for prefix in [self.name, self.module_name, self.alias]:
            value = self.__config.get("{}.{}".format(prefix, key), value)
        return value

    """Set a parameter for this module

    :param key: Name of the parameter to set
    :param value: New value of the parameter
    """

    def set(self, key, value):
        self.__config.set("{}.{}".format(self.name, key), value)

    """Override this method to define tasks that should be done during each
    update interval (for instance, querying an API, calling a CLI tool to get new
    date, etc.
    """

    def update(self):
        pass



    def update_wrapper(self):
        if self.background == True:
            if self.__thread and self.__thread.isAlive():
                return # skip this update interval
            self.__thread = threading.Thread(target=self.internal_update, args=(True,))
            self.__thread.start()
        else:
            self.internal_update(False)


    """Wrapper method that ensures that all exceptions thrown by the
    update() method are caught and displayed in a bumblebee_status.module.Error
    module
    """

    def internal_update(self, trigger_redraw=False):
        try:
            self.update()
            if trigger_redraw:
                core.event.trigger("update", [self.id], redraw_only=True)
        except Exception as e:
            self.set("interval", 1)
            module = Error(config=self.__config, module="error", error=str(e))
            self.__widgets = [module.widget()]
            self.update = module.update

    """Retrieves the list of widgets for this module

    :return: A list of widgets
    :rtype: list of bumblebee_status.widget.Widgets
    """

    def widgets(self):
        return self.__widgets

    """Removes all widgets of this module"""

    def clear_widgets(self):
        del self.__widgets[:]

    """Adds a widget to the module

    :param full_text: Text or callable (method) that defines the text of the widget (defaults to "")
    :param name: Name of the widget, defaults to None, which means autogenerate

    :return: The new widget
    :rtype: bumblebee_status.widget.Widget
    """

    def add_widget(self, full_text="", name=None):
        widget_id = "{}::{}".format(self.name, len(self.widgets()))
        widget = core.widget.Widget(full_text=full_text, name=name, widget_id=widget_id)
        self.widgets().append(widget)
        widget.module = self
        return widget

    """Convenience method to retrieve a named widget

    :param name: Name of widget to retrieve, defaults to None (in which case the first widget is returned)

    :return: The widget with the corresponding name, None if not found
    :rtype: bumblebee_status.widget.Widget
    """

    def widget(self, name=None, widget_id=None):
        if not name and not widget_id:
            return self.widgets()[0]

        for w in self.widgets():
            if name and w.name == name:
                return w
            if w.id == widget_id:
                return w
        return None

    """Override this method to define states for the module

    :param widget: Widget for which state should be returned

    :return: a list of states for this widget
    :rtype: list of strings
    """

    def state(self, widget):
        return []

    """Convenience method that sets warning and critical state for numbers

    :param value: Current value to calculate state against
    :param warn: Warning threshold
    :parm crit: Critical threshold

    :return: None if value is below both thresholds, "critical", "warning" as appropriate otherwise
    :rtype: string
    """

    def threshold_state(self, value, warn, crit):
        if value > float(self.parameter("critical", crit)):
            return "critical"
        if value > float(self.parameter("warning", warn)):
            return "warning"
        return None

    def register_callbacks(self):
        actions = [
            {"name": "left-click", "id": core.input.LEFT_MOUSE},
            {"name": "right-click", "id": core.input.RIGHT_MOUSE},
            {"name": "middle-click", "id": core.input.MIDDLE_MOUSE},
            {"name": "wheel-up", "id": core.input.WHEEL_UP},
            {"name": "wheel-down", "id": core.input.WHEEL_DOWN},
        ]
        for action in actions:
            if self.parameter(action["name"]):
                core.input.register(
                    self,
                    action["id"],
                    self.parameter(action["name"]),
                    util.format.asbool(
                        self.parameter("{}-wait".format(action["name"]), False)
                    ),
                )


class Error(Module):
    """Represents an "error" module

    :param module: The module name that produced the error
    :param error: The error message to display
    :param config: Configuration to apply to the module (defaults to an empty configuration)
    :param theme: Theme for this module, defaults to None, which means whatever is configured in "config"
    """

    def __init__(self, module, error, config=core.config.Config([]), theme=None):
        super().__init__(config, theme, core.widget.Widget(self.full_text))
        self.__module = module
        self.__error = error

    """Returns the error message
    :param widget: the error widget to display
    """

    def full_text(self, widget):
        return "{}: {}".format(self.__module, self.__error)

    """Overriden state, always returns critical (it *is* an error, after all"""

    def state(self, widget):
        return ["critical"]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Displays and changes the current keyboard layout

Requires the following executable:
    * xkb-switch

contributed by `somospocos <https://github.com/somospocos>`_ - many thanks!
"""

import core.module
import core.widget
import core.decorators
import core.input

import util.cli


class Module(core.module.Module):
    @core.decorators.every(seconds=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.current_layout))

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.__next_keymap)
        self.__current_layout = self.__get_current_layout()

    def current_layout(self, _):
        return self.__current_layout

    def __next_keymap(self, event):
        util.cli.execute("xkb-switch -n", ignore_errors=True)

    def __get_current_layout(self):
        try:
            res = util.cli.execute("xkb-switch")
            return res.split("\n")[0]
        except RuntimeError:
            return ["n/a"]

    def update(self):
        self.__current_layout = self.__get_current_layout()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Displays info about zpools present on the system

Parameters:
   * zpool.list: Comma-separated list of zpools to display info for. If empty, info for all zpools
     is displayed. (Default: '')
   * zpool.format: Format string, tags {name}, {used}, {left}, {size}, {percentfree}, {percentuse},
     {status}, {shortstatus}, {fragpercent}, {deduppercent} are supported.
     (Default: '{name} {used}/{size} ({percentfree}%)')
   * zpool.showio: Show also widgets detailing current read and write I/O (Default: true)
   * zpool.ioformat: Format string for I/O widget, tags {ops} (operations per seconds) and {band}
     (bandwidth) are supported. (Default: '{band}')
   * zpool.warnfree: Warn if free space is below this percentage (Default: 10)
   * zpool.sudo: Use sudo when calling the `zpool` binary. (Default: false)

Option `zpool.sudo` is intended for Linux users using zfsonlinux older than 0.7.0: In pre-0.7.0
releases of zfsonlinux regular users couldn't invoke even informative commands such as
`zpool list`. If this option is true, command `zpool list` is invoked with sudo. If this option
is used, the following (or ekvivalent) must be added to the `sudoers(5)`:

```
<username/ALL> ALL = (root) NOPASSWD: /usr/bin/zpool list
```

Be aware of security implications of doing this!

contributed by `adam-dej <https://github.com/adam-dej>`_ - many thanks!
"""

import time
import logging
from pkg_resources import parse_version

log = logging.getLogger(__name__)

import core.module

import util.cli
import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self._includelist = set(
            filter(
                lambda x: len(x) > 0,
                util.format.aslist(self.parameter("list", default="")),
            )
        )
        self._format = self.parameter(
            "format", default="{name} {shortstatus} {used}/{size} " + "({percentfree}%)"
        )
        self._usesudo = util.format.asbool(self.parameter("sudo", default=False))
        self._showio = util.format.asbool(self.parameter("showio", default=True))
        self._ioformat = self.parameter("ioformat", default="{band}")
        self._warnfree = int(self.parameter("warnfree", default=10))

    def update(self):
        self.clear_widgets()
        zfs_version_path = "/sys/module/zfs/version"
        # zpool list -H: List all zpools, use script mode (no headers and tabs as separators).
        try:
            with open(zfs_version_path, "r") as zfs_mod_version:
                zfs_version = zfs_mod_version.readline().rstrip().split("-")[0]
        except IOError:
            # ZFS isn't installed or the module isn't loaded, stub the version
            zfs_version = "0.0.0"
            logging.error(
                "ZFS version information not found at {}, check the module is loaded.".format(
                    zfs_version_path
                )
            )

        raw_zpools = util.cli.execute(
            ("sudo " if self._usesudo else "") + "zpool list -H"
        ).split("\n")

        for raw_zpool in raw_zpools:
            try:
                # Ignored fields (assigned to _) are 'expandsz' and 'altroot', also 'ckpoint' in ZFS 0.8.0+
                if parse_version(zfs_version) < parse_version("0.8.0"):
                    (
                        name,
                        size,
                        alloc,
                        free,
                        _,
                        frag,
                        cap,
                        dedup,
                        health,
                        _,
                    ) = raw_zpool.split("\t")
                else:
                    (
                        name,
                        size,
                        alloc,
                        free,
                        _,
                        _,
                        frag,
                        cap,
                        dedup,
                        health,
                        _,
                    ) = raw_zpool.split("\t")
                cap = cap.rstrip("%")
                percentuse = int(cap)
                percentfree = 100 - percentuse
                # There is a command, zpool iostat, which is however blocking and was therefore
                # causing issues.
                # Instead, we read file `/proc/spl/kstat/zfs/<poolname>/io` which contains
                # cumulative I/O statistics since boot (or pool creation). We store these values
                # (and timestamp) during each widget update, and during the next widget update we
                # use them to compute delta of transferred bytes, and using the last and current
                # timestamp the rate at which they have been transferred.
                with open("/proc/spl/kstat/zfs/{}/io".format(name), "r") as f:
                    # Third row provides data we need, we are interested in the first 4 values.
                    # More info about this file can be found here:
                    # https://github.com/zfsonlinux/zfs/blob/master/lib/libspl/include/sys/kstat.h#L580
                    # The 4 values are:
                    # nread, nwritten, reads, writes
                    iostat = list(map(int, f.readlines()[2].split()[:4]))
            except (ValueError, IOError):
                # Unable to parse info about this pool, skip it
                continue

            if self._includelist and name not in self._includelist:
                continue

            widget = self.widget(name)
            if not widget:
                widget = self.add_widget(name=name)
                widget.set("last_iostat", [0, 0, 0, 0])
                widget.set("last_timestamp", 0)

            delta_iostat = [b - a for a, b in zip(iostat, widget.get("last_iostat"))]
            widget.set("last_iostat", iostat)

            # From docs:
            #   > Note that even though the time is always returned as a floating point number, not
            #   > all systems provide time with a better precision than 1 second.
            # Be aware that that may affect the precision of reported I/O
            # Also, during one update cycle the reported I/O may be garbage if the system time
            # was changed.
            timestamp = time.time()
            delta_timestamp = widget.get("last_timestamp") - timestamp
            widget.set("last_timestamp", time.time())

            # abs is there because sometimes the result is -0
            rate_iostat = [abs(x / delta_timestamp) for x in delta_iostat]
            nread, nwritten, reads, writes = rate_iostat

            # theme.minwidth is not set since these values are not expected to change
            # rapidly
            widget.full_text(
                self._format.format(
                    name=name,
                    used=alloc,
                    left=free,
                    size=size,
                    percentfree=percentfree,
                    percentuse=percentuse,
                    status=health,
                    shortstatus=self._shortstatus(health),
                    fragpercent=frag,
                    deduppercent=dedup,
                )
            )
            widget.set("state", health)
            widget.set("percentfree", percentfree)
            widget.set("visited", True)

            if self._showio:
                wname, rname = [name + x for x in ["__write", "__read"]]
                widget_w = self.widget(wname)
                widget_r = self.widget(rname)
                if not widget_w or not widget_r:
                    widget_r = self.add_widget(name=rname)
                    widget_w = self.add_widget(name=wname)
                for w in [widget_r, widget_w]:
                    w.set(
                        "theme.minwidth",
                        self._ioformat.format(
                            ops=9999, band=util.format.bytefmt(999.99 * (1024 ** 2))
                        ),
                    )
                widget_w.full_text(
                    self._ioformat.format(
                        ops=round(writes), band=util.format.bytefmt(nwritten)
                    )
                )
                widget_r.full_text(
                    self._ioformat.format(
                        ops=round(reads), band=util.format.bytefmt(nread)
                    )
                )

    def state(self, widget):
        if widget.name.endswith("__read"):
            return "poolread"
        elif widget.name.endswith("__write"):
            return "poolwrite"

        state = widget.get("state")
        if state == "FAULTED":
            return [state, "critical"]
        elif state == "DEGRADED" or widget.get("percentfree") < self._warnfree:
            return [state, "warning"]

        return state

    @staticmethod
    def _shortstatus(status):
        # From `zpool(8)`, section Device Failure and Recovery:
        # A pool's health status is described by one of three states: online, degraded, or faulted.
        # An online pool has all devices operating normally.  A degraded pool is one in which one
        # or more devices have failed, but the data is still available due to a redundant
        # configuration.  A faulted pool has corrupted metadata, or one or more faulted devices, and
        # insufficient replicas to continue functioning.
        shortstate = {
            "DEGRADED": "DEG",
            "FAULTED": "FLT",
            "ONLINE": "ONL",
        }
        try:
            return shortstate[status]
        except KeyError:
            return ""


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the current song being played in DeaDBeeF and provides
some media control bindings.
Left click toggles pause, scroll up skips the current song, scroll
down returns to the previous song.

Requires the following library:
    * subprocess
Parameters:
    * deadbeef.format:    Format string (defaults to '{artist} - {title}')
      Available values are: {artist}, {title}, {album}, {length},
      {trackno}, {year}, {comment},
      {copyright}, {time}
      This is deprecated, but much simpler.
    * deadbeef.tf_format: A foobar2000 title formatting-style format string.
      These can be much more sophisticated than the standard
      format strings. This is off by default, but specifying
      any tf_format will enable it. If both deadbeef.format
      and deadbeef.tf_format are specified, deadbeef.tf_format
      takes priority.
    * deadbeef.tf_format_if_stopped: Controls whether or not the tf_format format
      string should be displayed even if no song is paused or
      playing. This could be useful if you want to implement
      your own stop strings with the built in logic. Any non-
      null value will enable this (by default the module will
      hide itself when the player is stopped).
    * deadbeef.previous:  Change binding for previous song (default is left click)
    * deadbeef.next:      Change binding for next song (default is right click)
    * deadbeef.pause:     Change binding for toggling pause (default is middle click)

    Available options for deadbeef.previous, deadbeef.next and deadbeef.pause are:
        LEFT_CLICK, RIGHT_CLICK, MIDDLE_CLICK, SCROLL_UP, SCROLL_DOWN

contributed by `joshbarrass <https://github.com/joshbarrass>`_ - many thanks!
"""
import sys

import subprocess
import logging

import core.module
import core.widget
import core.input
import core.decorators

import util.cli
import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.deadbeef))

        buttons = {
            "LEFT_CLICK": core.input.LEFT_MOUSE,
            "RIGHT_CLICK": core.input.RIGHT_MOUSE,
            "MIDDLE_CLICK": core.input.MIDDLE_MOUSE,
            "SCROLL_UP": core.input.WHEEL_UP,
            "SCROLL_DOWN": core.input.WHEEL_DOWN,
        }

        self._song = ""
        self._format = self.parameter("format", "{artist} - {title}")
        self._tf_format = self.parameter("tf_format", "")
        self._show_tf_when_stopped = util.format.asbool(
            self.parameter("tf_format_if_stopped", False)
        )
        prev_button = self.parameter("previous", "LEFT_CLICK")
        next_button = self.parameter("next", "RIGHT_CLICK")
        pause_button = self.parameter("pause", "MIDDLE_CLICK")

        self.now_playing = "deadbeef --nowplaying %a;%t;%b;%l;%n;%y;%c;%r;%e"
        self.now_playing_tf = "deadbeef --nowplaying-tf "
        cmd = "deadbeef "

        core.input.register(self, button=buttons[prev_button], cmd=cmd + "--prev")
        core.input.register(self, button=buttons[next_button], cmd=cmd + "--next")
        core.input.register(
            self, button=buttons[pause_button], cmd=cmd + "--play-pause"
        )

        # modify the tf_format if we don't want it to show on stop
        # this adds conditions to the query itself, rather than
        # polling to see if deadbeef is running
        # doing this reduces the number of calls we have to make
        if self._tf_format and not self._show_tf_when_stopped:
            self._tf_format = "$if($or(%isplaying%,%ispaused%),{query})".format(
                query=self._tf_format
            )

    @core.decorators.scrollable
    def deadbeef(self, widget):
        return self.string_song

    def hidden(self):
        return self.string_song == ""

    def update(self):
        widgets = self.widgets()
        try:
            if self._tf_format == "":  # no tf format set, use the old style
                return self.update_standard(widgets)
            return self.update_tf(widgets)
        except Exception as e:
            logging.exception(e)
            self._song = "error"

    def update_tf(self, widgets):
        ## ensure that deadbeef is actually running
        ## easiest way to do this is to check --nowplaying for
        ##  the string 'nothing'
        if util.cli.execute(self.now_playing) == "nothing":
            self._song = ""
            return
        ## perform the actual query -- these can be much more sophisticated
        data = util.cli.execute(self.now_playing_tf + self._tf_format)
        self._song = data

    def update_standard(self, widgets):
        data = util.cli.execute(self.now_playing)
        if data == "nothing":
            self._song = ""
        else:
            data = data.split(";")
            self._song = self._format.format(
                artist=data[0],
                title=data[1],
                album=data[2],
                length=data[3],
                trackno=data[4],
                year=data[5],
                comment=data[6],
                copyright=data[7],
                time=data[8],
            )

    @property
    def string_song(self):
        """\
Returns the current song as a string, either as a unicode() (Python <
3) or a regular str() (Python >= 3)
        """
        if sys.version_info.major < 3:
            return unicode(self._song)
        return str(self._song)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# pylint: disable=C0111,R0903

"""Displays the state of a Space API endpoint
Space API is an API for hackspaces based on JSON. See spaceapi.io for
an example.

Requires the following libraries:
    * requests
    * regex

Parameters:
    * spaceapi.url: String representation of the api endpoint
    * spaceapi.format: Format string for the output

Format Strings:
    * Format strings are indicated by double %%
    * They represent a leaf in the JSON tree, layers seperated by '.'
    * Boolean values can be overwritten by appending '%true%false'
      in the format string
    * Example: to reference 'open' in '{'state':{'open': true}}'
      you would write '%%state.open%%', if you also want
      to say 'Open/Closed' depending on the boolean you
      would write '%%state.open%Open%Closed%%'

contributed by `rad4day <https://github.com/rad4day>`_ - many thanks!
"""

import requests
import threading
import re
import json

import core.module
import core.widget
import core.input
import core.decorators


def formatStringBuilder(s, json):
    """
    Parses Format Strings
    Parameter:
        s -> format string
        json -> the spaceapi response object
    """
    identifiers = re.findall(r"%%.*?%%", s)
    for i in identifiers:
        ic = i[2:-2]  # Discard %%
        j = ic.split("%")

        # Only neither of, or both true AND false may be overwritten
        if len(j) != 3 and len(j) != 1:
            return "INVALID FORMAT STRING"

        if len(j) == 1:  # no overwrite
            s = s.replace(i, json[j[0]])
        elif json[j[0]]:  # overwrite for True
            s = s.replace(i, j[1])
        else:  # overwrite for False
            s = s.replace(i, j[2])
    return s


class Module(core.module.Module):
    @core.decorators.every(minutes=15)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.getState))

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.__forceReload)

        self.__data = {}
        self.__error = None
        self.__thread = None

        # The URL representing the api endpoint
        self.__url = self.parameter("url", default="http://club.entropia.de/spaceapi")
        self._format = self.parameter(
            "format", default=" %%space%%: %%state.open%Open%Closed%%"
        )

    def state(self, widget):
        try:
            if self.__error is not None:
                return ["critical"]
            elif self.__data["state.open"]:
                return ["warning"]
            else:
                return []
        except KeyError:
            return ["critical"]

    def update(self):
        if not self.__thread or self.__thread.is_alive() == False:
            self.__thread = threading.Thread(target=self.get_api_async, args=())
            self.__thread.start()

    def getState(self, widget):
        text = self._format
        if self.__error is not None:
            text = self.__error
        else:
            try:
                text = formatStringBuilder(self._format, self.__data)
            except KeyError:
                text = "KeyError"
        return text

    def get_api_async(self):
        try:
            with requests.get(self.__url, timeout=10) as request:
                # Can't implement error handling for python2.7 if I use
                # request.json() as it uses simplejson in newer versions
                self.__data = self.__flatten(json.loads(request.text))
                self.__error = None
        except requests.exceptions.Timeout:
            self.__error = "Timeout"
        except requests.exceptions.HTTPError:
            self.__error = "HTTP Error"
        except ValueError:
            self.__error = "Not a JSON response"
        core.event.trigger("update", [self.id], redraw_only=True)

    # left_mouse_button handler
    def __forceReload(self, event):
        if self.__thread:
            self.__thread.raise_exception()
        self.__error = "RELOADING"
        core.event.trigger("update", [self.id], redraw_only=True)

    # Flattens the JSON structure recursively, e.g. ['space']['open']
    # becomes ['space.open']
    def __flatten(self, json):
        out = {}
        for key in json:
            value = json[key]
            if type(value) is dict:
                flattened_key = self.__flatten(value)
                for fk in flattened_key:
                    out[key + "." + fk] = flattened_key[fk]
            else:
                out[key] = value
        return out


# Author: Tobias Manske <tobias@chaoswg.xyz>
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the current keyboard layout using libX11

Requires the following library:
    * libX11.so.6
and python module:
    * xkbgroup

Parameters:
    * layout-xkb.showname: Boolean that indicate whether the full name should be displayed. Defaults to false (only the symbol will be displayed)
    * layout-xkb.show_variant: Boolean that indecates whether the variant name should be displayed. Defaults to true.
"""

from xkbgroup import *

import logging

log = logging.getLogger(__name__)

import core.module
import core.widget
import core.input

import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.current_layout))

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.__next_keymap)
        core.input.register(self, button=core.input.RIGHT_MOUSE, cmd=self.__prev_keymap)
        self.__show_variant = util.format.asbool(self.parameter("show_variant", True))

    def __next_keymap(self, event):
        self.__set_keymap(1)

    def __prev_keymap(self, event):
        self.__set_keymap(-1)

    def __set_keymap(self, rotation):
        xkb = XKeyboard()
        if xkb.groups_count < 2:
            return  # nothing to do
        layouts = xkb.groups_symbols
        idx = layouts.index(xkb.group_symbol)
        xkb.group_symbol = str(layouts[(idx + rotation) % len(layouts)])

    def current_layout(self, widget):
        try:
            xkb = XKeyboard()
            log.debug("group num: {}".format(xkb.group_num))
            name = (
                xkb.group_name
                if util.format.asbool(self.parameter("showname"), False)
                else xkb.group_symbol
            )
            if self.__show_variant:
                return (
                    "{} ({})".format(name, xkb.group_variant)
                    if xkb.group_variant
                    else name
                )
            return name
        except Exception as e:
            print("got exception: {}".format(e))
            return "n/a"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Periodically checks the RTT of a configurable host using ICMP echos

Requires the following executable:
    * ping

Parameters:
    * ping.address : IP address to check
    * ping.timeout : Timeout for waiting for a reply (defaults to 5.0)
    * ping.probes  : Number of probes to send (defaults to 5)
    * ping.warning : Threshold for warning state, in seconds (defaults to 1.0)
    * ping.critical: Threshold for critical state, in seconds (defaults to 2.0)
"""

import re
import time

import core.module
import core.widget
import core.event
import core.decorators

import util.cli


class Module(core.module.Module):
    @core.decorators.every(seconds=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.rtt))

        self.background = True

        widget = self.widget()

        widget.set("address", self.parameter("address", "8.8.8.8"))
        widget.set("rtt-probes", self.parameter("probes", 5))
        widget.set("rtt-timeout", self.parameter("timeout", 5.0))
        widget.set("rtt-avg", 0.0)
        widget.set("rtt-unit", "")
        widget.set("packet-loss", 0)

    def rtt(self, widget):
        if widget.get("rtt-unreachable"):
            return "{}: unreachable".format(widget.get("address"))
        return "{}: {:.1f}{} ({}%)".format(
            widget.get("address"),
            widget.get("rtt-avg"),
            widget.get("rtt-unit"),
            widget.get("packet-loss"),
        )

    def state(self, widget):
        if widget.get("rtt-unreachable"):
            return ["critical"]
        return self.threshold_state(widget.get("rtt-avg"), 1000.0, 2000.0)

    def update(self):
        widget = self.widget()
        try:
            widget.set("rtt-unreachable", False)
            res = util.cli.execute(
                "ping -n -q -c {} -W {} {}".format(
                    widget.get("rtt-probes"),
                    widget.get("rtt-timeout"),
                    widget.get("address"),
                )
            )

            for line in res.split("\n"):
                if line.startswith(
                    "{} packets transmitted".format(widget.get("rtt-probes"))
                ):
                    m = re.search(r"(\d+)% packet loss", line)

                    widget.set("packet-loss", m.group(1))

                if not line.startswith("rtt"):
                    continue
                m = re.search(r"([0-9\.]+)/([0-9\.]+)/([0-9\.]+)/([0-9\.]+)\s+(\S+)", line)

                widget.set("rtt-min", float(m.group(1)))
                widget.set("rtt-avg", float(m.group(2)))
                widget.set("rtt-max", float(m.group(3)))
                widget.set("rtt-unit", m.group(5))
        except Exception as e:
            widget.set("rtt-unreachable", True)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# -*- coding: UTF-8 -*-

# smart function inspired by py-SMART https://github.com/freenas/py-SMART
# under Copyright (C) 2015 Marc Herndon and GPL2

"""Displays HDD smart status of different drives or all drives

Parameters:
    * smartstatus.display: how to display (defaults to 'combined', other choices: 'seperate' or 'singles')
    * smartstatus.drives: in the case of singles which drives to display, separated comma list value, multiple accepted (defaults to 'sda', example:'sda,sdc')
    * smartstatus.show_names: boolean in the form of "True" or "False" to show the name of the drives in the form of sda, sbd, combined or none at all. 
"""

import os

import shutil

import core.module
import core.decorators

import util.cli
import util.format


class Module(core.module.Module):
    @core.decorators.every(minutes=5)
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self.devices = self.list_devices()
        self.display = self.parameter("display", "combined")
        self.drives = self.parameter("drives", "sda")
        self.show_names = util.format.asbool(self.parameter("show_names", True))
        self.create_widgets()

    def create_widgets(self):
        if self.display == "combined":
            widget = self.add_widget()
            widget.set("device", "combined")
            widget.set("assessment", self.combined())
            self.output(widget)
        else:
            for device in self.devices:
                if self.display == "singles" and device not in self.drives:
                    continue
                widget = self.add_widget()
                widget.set("device", device)
                widget.set("assessment", self.smart(device))
                self.output(widget)

    def update(self):
        for widget in self.widgets():
            device = widget.get("device")
            if device == "combined":
                widget.set("assessment", self.combined())
                self.output(widget)
            else:
                widget.set("assessment", self.smart(device))
                self.output(widget)

    def output(self, widget):
        device = widget.get("device")
        assessment = widget.get("assessment")
        if self.show_names:
            widget.full_text("{}: {}".format(device, assessment))
        else:
            widget.full_text("{}".format(assessment))

    def state(self, widget):
        states = []
        assessment = widget.get("assessment")
        if assessment == "Pre-fail":
            states.append("warning")
        if assessment == "Fail":
            states.append("critical")
        return states

    def combined(self):
        for device in self.devices:
            result = self.smart(device)
            if result == "Fail":
                return "Fail"
            if result == "Pre-fail":
                return "Pre-fail"
        return "OK"

    def list_devices(self):
        for (root, folders, files) in os.walk("/dev"):
            if root == "/dev":
                devices = {
                    "".join(filter(lambda i: i.isdigit() == False, file))
                    for file in files
                    if "sd" in file
                }
                nvme = {
                    file for file in files if ("nvme0n" in file and "p" not in file)
                }
                devices.update(nvme)
                return devices

    def smart(self, disk_name):
        smartctl = shutil.which("smartctl")
        assessment = None

        output = util.cli.execute(
            "sudo {} --health {}".format(smartctl, os.path.join("/dev/", disk_name))
        )
        output = output.split("\n")
        line = output[4]
        if "SMART" in line:
            if any([i in line for i in ["PASSED", "OK"]]):
                assessment = "OK"
            else:
                assessment = "Fail"

        if assessment == "OK":
            output = util.cli.execute(
                "sudo {} -A {}".format(smartctl, os.path.join("/dev/", disk_name))
            )
            output = output.split("\n")
            for line in output:
                if "Pre-fail" in line:
                    assessment = "Pre-fail"
        return assessment


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Shows yubikey information

Requires: https://github.com/Yubico/python-yubico

The output indicates that a YubiKey is not connected or it displays
the corresponding serial number.


contributed by `EmmaTinten <https://github.com/EmmaTinten>`_ - many thanks!
"""

import yubico

import core.module
import core.widget
import core.decorators


class Module(core.module.Module):
    @core.decorators.every(seconds=5)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.keystate))
        self.__keystate = "No YubiKey"

    def keystate(self, widget):
        return self.__keystate

    def update(self):
        try:
            self.__keystate = "YubiKey: " + str(
                yubico.find_yubikey(debug=False).serial()
            )
        except yubico.yubico_exception.YubicoError:
            self.__keystate = "No YubiKey"
        except Exception:
            self.__keystate = "n/a"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest
import types

import core.theme
import core.event
import core.widget
import core.module


class SampleModule(core.module.Module):
    def __init__(self, widgets, config=core.config.Config([]), theme=None):
        super().__init__(config, theme, widgets)
        self.name = "test"


@pytest.fixture(autouse=True)
def clear_events():
    core.event.clear()


@pytest.fixture
def defaultsTheme():
    return {"defaults": {"fg": "red", "bg": "black"}}


@pytest.fixture
def cycleTheme():
    return {
        "cycle": [
            {"fg": "red", "bg": "black"},
            {"fg": "black", "bg": "red"},
            {"fg": "white", "bg": "blue"},
        ]
    }


@pytest.fixture
def colorTheme():
    return {"colors": [{"red": "#ff0000", "blue": "#0000ff"}]}


@pytest.fixture
def walTheme():
    return {"colors": ["wal"]}


@pytest.fixture
def cycleValueTheme():
    return {"defaults": {"fg": ["red", "green", "blue"]}}


@pytest.fixture
def stateTheme():
    return {"warning": {"fg": "yellow"}, "critical": {"fg": "red"}}


@pytest.fixture
def overlayTheme():
    return {
        "load": {"prefix": "a"},
        "test": {"load": {"prefix": "b"}, "prefix": "x"},
    }


def test_invalid_theme():
    with pytest.raises(RuntimeError):
        core.theme.Theme("this-theme-does-not-exist")


def test_valid_theme():
    theme = core.theme.Theme("default")
    assert theme.name == "default"


def test_defaults(defaultsTheme):
    theme = core.theme.Theme(raw_data=defaultsTheme)

    assert theme.get("fg") == defaultsTheme["defaults"]["fg"]
    assert theme.get("bg") == defaultsTheme["defaults"]["bg"]


def test_cycle(mocker, cycleTheme):
    theme = core.theme.Theme(raw_data=cycleTheme)

    assert theme.get("bg", "previous") == None
    assert theme.get("fg") == cycleTheme["cycle"][0]["fg"]
    assert theme.get("bg") == cycleTheme["cycle"][0]["bg"]

    core.event.trigger("next-widget")

    assert theme.get("bg", "previous") == cycleTheme["cycle"][0]["bg"]

    core.event.trigger("next-widget")

    assert theme.get("fg") == cycleTheme["cycle"][2]["fg"]
    assert theme.get("bg") == cycleTheme["cycle"][2]["bg"]

    stdout = mocker.patch("core.output.sys.stdout")

    core.event.trigger("draw")

    assert theme.get("fg") == cycleTheme["cycle"][0]["fg"]
    assert theme.get("bg") == cycleTheme["cycle"][0]["bg"]


def test_custom_iconset(defaultsTheme):
    theme = core.theme.Theme(raw_data=defaultsTheme)

    assert theme.get("padding") != "aaa"
    assert theme.get("fg") != "blue"

    theme = core.theme.Theme(
        raw_data=defaultsTheme, iconset={"defaults": {"padding": "aaa", "fg": "blue"}}
    )

    assert theme.get("padding") == "aaa"
    assert theme.get("fg") == "blue"  # test override


def test_colors(defaultsTheme, colorTheme):
    theme = core.theme.Theme(raw_data=defaultsTheme)
    assert theme.keywords() == {}

    theme = core.theme.Theme(raw_data=colorTheme)
    assert theme.keywords() == colorTheme["colors"][0]


def test_wal_colors(mocker, walTheme):
    io = mocker.patch("core.theme.io")
    os = mocker.patch("core.theme.os")

    os.path.isfile.return_value = True
    io.open.return_value = mocker.MagicMock()
    io.open.return_value.__enter__.return_value.read.return_value = """
        { "colors": { "red": "#ff0000" } }
    """

    theme = core.theme.Theme(raw_data=walTheme)

    assert theme.keywords() == {"red": "#ff0000"}


def test_wal_special(mocker, walTheme):
    io = mocker.patch("core.theme.io")
    os = mocker.patch("core.theme.os")

    os.path.isfile.return_value = True
    io.open.return_value.__enter__.return_value.read.return_value = """
        { "special": { "background": "#ff0000" } }
    """

    theme = core.theme.Theme(raw_data=walTheme)

    assert theme.keywords() == {"background": "#ff0000"}


def test_cycle_value(cycleValueTheme):
    widget = core.widget.Widget()
    expected = cycleValueTheme["defaults"]["fg"]
    theme = core.theme.Theme(raw_data=cycleValueTheme)

    for i in range(0, len(expected) * 3):
        assert theme.get("fg", widget) == expected[i % len(expected)]
        # ensure multiple invocations are OK
        assert theme.get("fg", widget) == expected[i % len(expected)]
        core.event.trigger("draw")


def test_state(stateTheme):
    widget = core.widget.Widget()
    theme = core.theme.Theme(raw_data=stateTheme)

    assert theme.get("fg", widget) == None

    widget.state = types.MethodType(lambda self: ["warning"], widget)
    assert theme.get("fg", widget) == stateTheme["warning"]["fg"]

    widget.state = types.MethodType(lambda self: ["critical"], widget)
    assert theme.get("fg", widget) == stateTheme["critical"]["fg"]


def test_overlay(overlayTheme):
    widget = core.widget.Widget()
    module = SampleModule(widget)
    theme = core.theme.Theme(raw_data=overlayTheme)

    assert theme.get("prefix", widget) == overlayTheme[module.name]["prefix"]

    widget.state = types.MethodType(lambda self: ["load"], widget)

    assert theme.get("prefix", widget) == overlayTheme[module.name]["load"]["prefix"]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Shows free diskspace, total diskspace and the percentage of free disk space.

Parameters:
    * disk.warning: Warning threshold in % of disk space (defaults to 80%)
    * disk.critical: Critical threshold in % of disk space (defaults ot 90%)
    * disk.path: Path to calculate disk usage from (defaults to /)
    * disk.open: Which application / file manager to launch (default xdg-open)
    * disk.format: Format string, tags {path}, {used}, {left}, {size} and {percent} (defaults to '{path} {used}/{size} ({percent:05.02f}%)')
"""

import os

import core.module
import core.widget
import core.input

import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.diskspace))

        self._path = self.parameter("path", "/")
        self._format = self.parameter("format", "{used}/{size} ({percent:05.02f}%)")

        self._used = 0
        self._left = 0
        self._size = 0
        self._percent = 0

        core.input.register(
            self,
            button=core.input.LEFT_MOUSE,
            cmd="{} {}".format(self.parameter("open", "xdg-open"), self._path),
        )

    def diskspace(self, widget):
        used_str = util.format.byte(self._used)
        size_str = util.format.byte(self._size)
        left_str = util.format.byte(self._left)
        percent_str = self._percent

        return self._format.format(
            path=self._path,
            used=used_str,
            left=left_str,
            size=size_str,
            percent=percent_str,
        )

    def update(self):
        st = os.statvfs(self._path)
        self._size = st.f_blocks * st.f_frsize
        self._used = (st.f_blocks - st.f_bfree) * st.f_frsize
        self._left = self._size - self._used
        self._percent = 100.0 * self._used / self._size

    def state(self, widget):
        return self.threshold_state(self._percent, 80, 90)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays battery status, remaining percentage and charging information.

Parameters:
    * battery.device              : Comma-separated list of battery devices to read information from (defaults to auto for auto-detection)
    * battery.warning             : Warning threshold in % of remaining charge (defaults to 20)
    * battery.critical            : Critical threshold in % of remaining charge (defaults to 10)
    * battery.showdevice          : If set to 'true', add the device name to the widget (defaults to False)
    * battery.decorate            : If set to 'false', hides additional icons (charging, etc.) (defaults to True)
    * battery.showpowerconsumption: If set to 'true', show current power consumption (defaults to False)
    * battery.compact-devices     : If set to 'true', compacts multiple batteries into a single entry (default to False)

(partially) contributed by `martindoublem <https://github.com/martindoublem>`_ - many thanks!
"""

import os
import glob
import logging

log = logging.getLogger(__name__)

try:
    import power
except ImportError:
    log.warning('unable to import module "power": Time estimates will not be available')

import core.module
import core.input

import util.format


class BatteryManager(object):
    def remaining(self):
        try:
            estimate = power.PowerManagement().get_time_remaining_estimate()
            # do not show remaining if on AC
            if estimate == power.common.TIME_REMAINING_UNLIMITED:
                return None
            return estimate * 60  # return value in seconds
        except Exception as e:
            return -1
        return -1

    def read(self, battery, component, default=None):
        path = "/sys/class/power_supply/{}".format(battery)
        if not os.path.exists(path):
            return default
        try:
            with open("{}/{}".format(path, component)) as f:
                return f.read().strip()
        except IOError:
            return "n/a"
        return default

    def capacity(self, battery):
        capacity = self.read(battery, "capacity", 100)
        if capacity != "n/a":
            capacity = int(capacity)

        return capacity if capacity < 100 else 100

    def capacity_all(self, batteries):
        now = 0
        full = 0
        for battery in batteries:
            try:
                with open(
                    "/sys/class/power_supply/{}/energy_full".format(battery)
                ) as f:
                    full += int(f.read())
                with open("/sys/class/power_supply/{}/energy_now".format(battery)) as f:
                    now += int(f.read())
            except IOError:
                return "n/a"
        return int(float(now) / float(full) * 100.0)

    def isac(self, battery):
        path = "/sys/class/power_supply/{}".format(battery)
        return not os.path.exists(path)

    def isac_any(self, batteries):
        for battery in batteries:
            if self.isac(battery):
                return True
        return False

    def consumption(self, battery):
        consumption = self.read(battery, "power_now", "n/a")
        if consumption == "n/a":
            return "n/a"
        return "{}W".format(int(consumption) / 1000000)

    def charge(self, battery):
        return self.read(battery, "status", "n/a")

    def charge_any(self, batteries):
        for battery in batteries:
            if self.charge(battery) == "Discharging":
                return "Discharging"
        return "Charged"


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self.__manager = BatteryManager()

        self._batteries = util.format.aslist(self.parameter("device", "auto"))
        if self._batteries[0] == "auto":
            self._batteries = [
                os.path.basename(battery)
                for battery in glob.glob("/sys/class/power_supply/BAT*")
            ]
        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd="gnome-power-statistics"
        )

        if len(self._batteries) == 0:
            widget = self.add_widget(full_text=self.ac, name="ac")
            widget.set("ac", True)
            widget.set("capacity", 100)
        else:
            if util.format.asbool(self.parameter("compact-devices", False)):
                widget = self.add_widget(full_text=self.capacity, name="all-batteries")
            else:
                for battery in self._batteries:
                    log.debug("adding new widget for {}".format(battery))
                    widget = self.add_widget(full_text=self.capacity, name=battery)

        for w in self.widgets():
            if util.format.asbool(self.parameter("decorate", True)) == False:
                widget.set("theme.exclude", "suffix")

    def ac(self, widget):
        return "ac"

    def capacity(self, widget):
        if widget.name == "all-batteries":
            capacity = self.__manager.capacity_all(self._batteries)
        else:
            capacity = self.__manager.capacity(widget.name)
        widget.set("capacity", capacity)
        widget.set("ac", self.__manager.isac_any(self._batteries))
        widget.set("theme.minwidth", "100%")

        # Read power conumption
        if util.format.asbool(self.parameter("showpowerconsumption", False)):
            output = "{}% ({})".format(
                capacity, self.__manager.consumption(widget.name)
            )
        else:
            output = "{}%".format(capacity)

        if (
            util.format.asbool(self.parameter("showremaining", True))
            and self.__manager.charge(widget.name) == "Discharging"
        ):
            remaining = self.__manager.remaining()
            if remaining >= 0:
                output = "{} {}".format(
                    output, util.format.duration(remaining, compact=True, unit=True)
                )

        if util.format.asbool(self.parameter("showdevice", False)):
            output = "{} ({})".format(output, widget.name)

        return output

    def state(self, widget):
        state = []
        capacity = widget.get("capacity")

        if capacity < 0:
            log.debug("battery state: {}".format(state))
            return ["critical", "unknown"]

        if capacity < int(self.parameter("critical", 10)):
            state.append("critical")
        elif capacity < int(self.parameter("warning", 20)):
            state.append("warning")

        if widget.get("ac"):
            state.append("AC")
        else:
            if widget.name == "all-batteries":
                charge = self.__manager.charge_any(self._batteries)
            else:
                charge = self.__manager.charge(widget.name)
            if charge == "Discharging":
                state.append(
                    "discharging-{}".format(
                        min([10, 25, 50, 80, 100], key=lambda i: abs(i - capacity))
                    )
                )
            elif charge == "Unknown":
                state.append(
                    "unknown-{}".format(
                        min([10, 25, 50, 80, 100], key=lambda i: abs(i - capacity))
                    )
                )
            else:
                if capacity > 95:
                    state.append("charged")
                else:
                    state.append("charging")
        return state


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# -*- coding: utf-8 -*-
# pylint: disable=C0111,R0903

""" system module

adds the possibility to
	* shutdown
	* reboot

the system.
	
Per default a confirmation dialog is shown before the actual action is performed.
	
Parameters:
	* system.confirm: show confirmation dialog before performing any action (default: true) 
        * system.reboot: specify a reboot command (defaults to 'reboot')
        * system.shutdown: specify a shutdown command (defaults to 'shutdown -h now')
        * system.logout: specify a logout command (defaults to 'i3exit logout')
        * system.switch_user: specify a command for switching the user (defaults to 'i3exit switch_user')
        * system.lock: specify a command for locking the screen (defaults to 'i3exit lock')
        * system.suspend: specify a command for suspending (defaults to 'i3exit suspend')
        * system.hibernate: specify a command for hibernating (defaults to 'i3exit hibernate')

contributed by `bbernhard <https://github.com/bbernhard>`_ - many thanks!
"""

import logging
import functools

try:
    import tkinter as tk
    from tkinter import messagebox as tkmessagebox
except ImportError:
    logging.warning("failed to import tkinter - bumblebee popups won't work!")

import core.module
import core.widget
import core.input
import core.decorators

import util.cli
import util.popup
import util.format


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.text))

        self.__confirm = util.format.asbool(self.parameter("confirm", True))

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.popup)

    def text(self, widget):
        return ""

    def __on_command(self, header, text, command):
        do_it = True
        if self.__confirm:
            root = tk.Tk()
            root.withdraw()
            root.focus_set()

            do_it = tkmessagebox.askyesno(header, text)
            root.destroy()

        if do_it:
            util.cli.execute(command)

    def popup(self, widget):
        menu = util.popup.menu()
        reboot_cmd = self.parameter("reboot", "reboot")
        shutdown_cmd = self.parameter("shutdown", "shutdown -h now")
        logout_cmd = self.parameter("logout", "i3exit logout")
        switch_user_cmd = self.parameter("switch_user", "i3exit switch_user")
        lock_cmd = self.parameter("lock", "i3exit lock")
        suspend_cmd = self.parameter("suspend", "i3exit suspend")
        hibernate_cmd = self.parameter("hibernate", "i3exit hibernate")

        menu.add_menuitem(
            "shutdown",
            callback=functools.partial(
                self.__on_command, "Shutdown", "Shutdown?", shutdown_cmd
            ),
        )
        menu.add_menuitem(
            "reboot",
            callback=functools.partial(
                self.__on_command, "Reboot", "Reboot?", reboot_cmd
            ),
        )
        menu.add_menuitem(
            "log out",
            callback=functools.partial(
                self.__on_command, "Log out", "Log out?", "i3exit logout"
            ),
        )
        # don't ask for these
        menu.add_menuitem(
            "switch user", callback=functools.partial(util.cli.execute, switch_user_cmd)
        )
        menu.add_menuitem(
            "lock", callback=functools.partial(util.cli.execute, lock_cmd)
        )
        menu.add_menuitem(
            "suspend", callback=functools.partial(util.cli.execute, suspend_cmd)
        )
        menu.add_menuitem(
            "hibernate", callback=functools.partial(util.cli.execute, hibernate_cmd)
        )

        menu.show(widget, 0, 0)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Print the branch and git status for the
currently focused window.

Requires:
    * xcwd
    * Python module 'pygit2'
"""

import os
import pygit2

import core.module

import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self.__error = False

    def hidden(self):
        return self.__error

    def update(self):
        state = {}
        self.clear_widgets()
        try:
            directory = util.cli.execute("xcwd").strip()
            directory = self.__get_git_root(directory)
            repo = pygit2.Repository(directory)

            self.add_widget(name="git.main", full_text=repo.head.shorthand)

            for filepath, flags in repo.status().items():
                if (
                    flags == pygit2.GIT_STATUS_WT_NEW
                    or flags == pygit2.GIT_STATUS_INDEX_NEW
                ):
                    state["new"] = True
                if (
                    flags == pygit2.GIT_STATUS_WT_DELETED
                    or flags == pygit2.GIT_STATUS_INDEX_DELETED
                ):
                    state["deleted"] = True
                if (
                    flags == pygit2.GIT_STATUS_WT_MODIFIED
                    or flags == pygit2.GIT_STATUS_INDEX_MODIFIED
                ):
                    state["modified"] = True
            self.__error = False
            if "new" in state:
                self.add_widget(name="git.new")
            if "modified" in state:
                self.add_widget(name="git.modified")
            if "deleted" in state:
                self.add_widget(name="git.deleted")

        except Exception as e:
            self.__error = True

    def state(self, widget):
        return widget.name.split(".")[1]

    def __get_git_root(self, directory):
        while len(directory) > 1:
            if os.path.exists(os.path.join(directory, ".git")):
                return directory
            directory = "/".join(directory.split("/")[0:-1])
        return "/"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

import sys
import shlex

import core.module
import core.widget
import core.config
import core.input


@pytest.fixture(autouse=True)
def clear_events():
    core.event.clear()


@pytest.fixture
def empty_config():
    return core.config.Config([])


@pytest.fixture
def widget_a():
    return core.widget.Widget("randomwWidget content", name="A")


@pytest.fixture
def widget_b():
    return core.widget.Widget("another randomwWidget content", name="B")


class SampleModule(core.module.Module):
    def update(self):
        if self.fail:
            raise Exception(self.error)
        pass


def test_loadinvalid_module(mocker):
    config = mocker.MagicMock()
    module = core.module.load(module_name="i-do-not-exist", config=config)
    assert module.__class__.__module__ == "core.module"
    assert module.__class__.__name__ == "Error"


@pytest.mark.skipif(
    sys.version_info.major == 3 and sys.version_info.minor in [4, 5],
    reason="importlib error reporting in Python 3.{4,5} different",
)
def test_importerror(mocker):
    importlib = mocker.patch("core.module.importlib")
    importlib.import_module.side_effect = ImportError("some-error")
    config = mocker.MagicMock()

    module = core.module.load(module_name="test", config=config)

    assert module.__class__.__name__ == "Error"
    assert module.widget().full_text() == "test: some-error"


def test_loadvalid_module():
    module = core.module.load(module_name="test")
    assert module.__class__.__module__ == "modules.core.test"
    assert module.__class__.__name__ == "Module"
    assert module.state(None) == []


def test_empty_widgets():
    module = core.module.Module(widgets=[])
    assert module.widgets() == []


def test_error_widget():
    cfg = core.config.Config(shlex.split("-p test_module.foo=5"))
    module = core.module.Error("test-mod", "xyz", config=cfg)
    full_text = module.full_text(module.widget())

    assert module.state(None) == ["critical"]
    assert "test-mod" in full_text
    assert "xyz" in full_text


def test_single_widget(widget_a):
    module = core.module.Module(widgets=widget_a)
    assert module.widgets() == [widget_a]


def test_widget_list(widget_a, widget_b):
    module = core.module.Module(widgets=[widget_a, widget_b])
    assert module.widgets() == [widget_a, widget_b]


def test_module_Name():
    module = SampleModule()
    assert module.name == "test_module"
    assert module.module_name == "test_module"


def testvalid_parameter():
    cfg = core.config.Config(shlex.split("-p test_module.foo=5"))
    module = SampleModule(config=cfg)
    assert module.parameter("foo") == "5"


def test_default_parameter(empty_config):
    module = SampleModule(config=empty_config)
    assert module.parameter("foo", "default") == "default"


def test_default_is_none(empty_config):
    module = SampleModule(config=empty_config)
    assert module.parameter("foo") == None


def test_error_widget(empty_config):
    module = SampleModule(config=empty_config)
    module.fail = True
    module.error = "!!"
    module.update_wrapper()
    assert len(module.widgets()) == 1
    assert module.widget().full_text() == "error: !!"


def test_get_widget_by_name(empty_config, widget_a, widget_b):
    module = SampleModule(config=empty_config, widgets=[widget_a, widget_b])

    assert module.widget(widget_a.name) == widget_a
    assert module.widget(widget_b.name) == widget_b
    assert module.widget("i-do-not-exist") == None
    assert module.widget() == widget_a

def test_get_widget_by_id(empty_config, widget_a, widget_b):
    module = SampleModule(config=empty_config, widgets=[widget_a, widget_b])

    assert module.widget(widget_id=widget_a.id) == widget_a
    assert module.widget(widget_id=widget_b.id) == widget_b
    assert module.widget(widget_id="i-do-not-exist") == None


def test_default_thresholds(empty_config, widget_a, widget_b):
    module = SampleModule(config=empty_config, widgets=[widget_a, widget_b])

    assert module.threshold_state(100, 80, 99) == "critical"
    assert module.threshold_state(100, 80, 100) == "warning"
    assert module.threshold_state(81, 80, 100) == "warning"
    assert module.threshold_state(80, 80, 100) == None
    assert module.threshold_state(10, 80, 100) == None


def test_configured_callbacks(mocker, empty_config, widget_a, widget_b):
    module = SampleModule(config=empty_config, widgets=[widget_a, widget_b])

    cmd = "sample-tool arg1 arg2 arg3"
    module.set("left-click", cmd)
    module.register_callbacks()

    cli = mocker.patch("core.input.util.cli")
    cli.execute.return_value = ""
    core.input.trigger(
        {"button": core.input.LEFT_MOUSE, "instance": module.id,}
    )

    cli.execute.assert_called_once_with(cmd, wait=False, shell=True)


def test_configured_callbacks_with_parameters(mocker, empty_config, widget_a):
    module = SampleModule(config=empty_config, widgets=[widget_a])

    cmd = "sample-tool {instance} {name}"
    module.set("left-click", cmd)
    module.register_callbacks()

    cli = mocker.patch("core.input.util.cli")
    cli.execute.return_value = ""
    core.input.trigger(
        {"button": core.input.LEFT_MOUSE, "instance": module.id, "name": "sample-name",}
    )

    cli.execute.assert_called_once_with(
        cmd.format(instance=module.id, name="sample-name"), wait=False, shell=True,
    )


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""
Show progress for cp, mv, dd, ...

Parameters:
   * progress.placeholder: Text to display while no process is running (defaults to 'n/a')
   * progress.barwidth: Width of the progressbar if it is used (defaults to 8)
   * progress.format: Format string (defaults to '{bar} {cmd} {arg}')
     Available values are: {bar} {pid} {cmd} {arg} {percentage} {quantity} {speed} {time}
   * progress.barfilledchar: Character used to draw the filled part of the bar (defaults to '#'), notice that it can be a string
   * progress.baremptychar: Character used to draw the empty part of the bar (defaults to '-'), notice that it can be a string

Requires the following executable:
   * progress

contributed by `remi-dupre <https://github.com/remi-dupre>`_ - many thanks!
"""

import core.module
import core.widget

import util.cli
import util.format

import re


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.get_progress_text))
        self.__active = False

    def get_progress_text(self, widget):
        if self.update_progress_info(widget):
            width = util.format.asint(self.parameter("barwidth", 8))
            count = round((width * widget.get("per")) / 100)
            filledchar = self.parameter("barfilledchar", "#")
            emptychar = self.parameter("baremptychar", "-")

            bar = "[{}{}]".format(filledchar * count, emptychar * (width - count))

            str_format = self.parameter("format", "{bar} {cmd} {arg}")
            return str_format.format(
                bar=bar,
                pid=widget.get("pid"),
                cmd=widget.get("cmd"),
                arg=widget.get("arg"),
                percentage=widget.get("per"),
                quantity=widget.get("qty"),
                speed=widget.get("spd"),
                time=widget.get("tim"),
            )
        else:
            return self.parameter("placeholder", "n/a")

    def update_progress_info(self, widget):
        """Update widget's informations about the copy"""
        if not self.__active:
            return

        # These regex extracts following groups:
        #  1. pid
        #  2. command
        #  3. arguments
        #  4. progress (xx.x formated)
        #  5. quantity (.. unit / .. unit formated)
        #  6. speed
        #  7. time remaining
        extract_nospeed = re.compile(
            r"\[ *(\d*)\] ([a-zA-Z]*) (.*)\n\t(\d*\.*\d*)% \((.*)\)\n.*"
        )
        extract_wtspeed = re.compile(
            r"\[ *(\d*)\] ([a-zA-Z]*) (.*)\n\t(\d*\.*\d*)% \((.*)\) (\d*\.\d .*) remaining (\d*:\d*:\d*)\n.*"
        )

        try:
            raw = util.cli.execute("progress -qW 0.1")
            result = extract_wtspeed.match(raw)

            if not result:
                # Abord speed measures
                raw = util.cli.execute("progress -q")
                result = extract_nospeed.match(raw)

                widget.set("spd", "???.? B/s")
                widget.set("tim", "??:??:??")
            else:
                widget.set("spd", result.group(6))
                widget.set("tim", result.group(7))

            widget.set("pid", int(result.group(1)))
            widget.set("cmd", result.group(2))
            widget.set("arg", result.group(3))
            widget.set("per", float(result.group(4)))
            widget.set("qty", result.group(5))
            return True
        except Exception:
            return False

    def update(self):
        self.__active = bool(util.cli.execute("progress -q"))

    def state(self, widget):
        if self.__active:
            return "copying"
        return "pending"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the pi-hole status (up/down) together with the number of ads that were blocked today

Parameters:
    * pihole.address     : pi-hole address (e.q: http://192.168.1.3)
    * pihole.pwhash      : pi-hole webinterface password hash (can be obtained from the /etc/pihole/SetupVars.conf file)

contributed by `bbernhard <https://github.com/bbernhard>`_ - many thanks!
"""

import requests

import core.module
import core.widget
import core.input


class Module(core.module.Module):
    @core.decorators.every(minutes=1)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.pihole_status))

        self._pihole_address = self.parameter("address", "")
        self._pihole_pw_hash = self.parameter("pwhash", "")
        self._pihole_status = None
        self._ads_blocked_today = "-"
        self.update_pihole_status()

        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd=self.toggle_pihole_status
        )

    def pihole_status(self, widget):
        if self._pihole_status is None:
            return "pi-hole unknown"
        return "pi-hole {}".format(
            "up {} blocked".format(self._ads_blocked_today)
            if self._pihole_status
            else "down"
        )

    def update_pihole_status(self):
        try:
            data = requests.get(self._pihole_address + "/admin/api.php?summary").json()
            self._pihole_status = True if data["status"] == "enabled" else False
            self._ads_blocked_today = data["ads_blocked_today"]
        except Exception as e:
            self._pihole_status = None

    def toggle_pihole_status(self, widget):
        if self._pihole_status is not None:
            try:
                req = None
                if self._pihole_status:
                    req = requests.get(
                        self._pihole_address
                        + "/admin/api.php?disable&auth="
                        + self._pihole_pw_hash
                    )
                else:
                    req = requests.get(
                        self._pihole_address
                        + "/admin/api.php?enable&auth="
                        + self._pihole_pw_hash
                    )
                if req is not None:
                    if req.status_code == 200:
                        status = req.json()["status"]
                        self._pihole_status = False if status == "disabled" else True
            except:
                pass

    def update(self):
        self.update_pihole_status()

    def state(self, widget):
        if self._pihole_status is None:
            return []
        elif self._pihole_status:
            return ["enabled"]
        return ["disabled", "warning"]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the current color temperature of redshift

Requires the following executable:
    * redshift

Parameters:
    * redshift.location : location provider, either of 'auto' (default), 'geoclue2',
      'ipinfo' or 'manual'
      'auto' uses whatever redshift is configured to do
    * redshift.lat : latitude if location is set to 'manual'
    * redshift.lon : longitude if location is set to 'manual'
    * redshift.show_transition: information about the transitions (x% day) defaults to True
"""

import re
import threading

import core.module
import core.widget
import core.input
import core.decorators

import util.cli
import util.format
import util.location


def get_redshift_value(module):
    widget = module.widget()
    location = module.parameter("location", "auto")
    lat = module.parameter("lat", None)
    lon = module.parameter("lon", None)

    # Even if location method is set to manual, if we have no lat or lon,
    # fall back to the geoclue2 method.
    if location == "manual" and (lat is None or lon is None):
        location = "geoclue2"

    command = ["redshift", "-p"]
    if location == "manual":
        command.extend(["-l", "{}:{}".format(lat, lon)])
    if location == "geoclue2":
        command.extend(["-l", "geoclue2"])

    try:
        res = util.cli.execute(" ".join(command))
    except Exception:
        res = ""
    widget.set("temp", "n/a")
    widget.set("transition", "")
    widget.set("state", "day")
    for line in res.split("\n"):
        line = line.lower()
        if "temperature" in line:
            widget.set("temp", line.split(" ")[2])
        if "period" in line:
            state = line.split(" ")[1]
            if "day" in state:
                widget.set("state", "day")
            elif "night" in state:
                widget.set("state", "night")
            else:
                widget.set("state", "transition")
                match = re.search(r"(\d+)\.\d+% ([a-z]+)", line)
                widget.set(
                    "transition", "({}% {})".format(match.group(1), match.group(2))
                )
    core.event.trigger("update", [widget.module.id], redraw_only=True)


class Module(core.module.Module):
    @core.decorators.every(seconds=10)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.text))

        self.__thread = None
        self.show_transition = util.format.asbool(
            self.parameter("show_transition", True)
        )

        if self.parameter("location", "") == "ipinfo":
            # override lon/lat with ipinfo
            try:
                location = util.location.coordinates()
                self.set("lat", location[0])
                self.set("lon", location[1])
                self.set("location", "manual")
            except Exception:
                # Fall back to geoclue2.
                self.set("location", "geoclue2")

        self._text = ""

    def text(self, widget):
        val = widget.get("temp", "n/a")
        transition = widget.get("transition", "")
        if transition and self.show_transition:
            val = "{} {}".format(val, transition)
        return val

    def update(self):
        if self.__thread is not None and self.__thread.isAlive():
            return
        self.__thread = threading.Thread(target=get_redshift_value, args=(self,))
        self.__thread.start()

    def state(self, widget):
        return widget.get("state", None)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays and changes the current selected prime video card

Left click will call 'sudo prime-select nvidia'
Right click will call 'sudo prime-select nvidia'

Running these commands without a password requires editing your sudoers file
(always use visudo, it's very easy to make a mistake and get locked out of your computer!)

sudo visudo -f /etc/sudoers.d/prime

Then put a line like this in there:

    user    ALL=(ALL) NOPASSWD: /usr/bin/prime-select

If you can't figure out the sudoers thing, then don't worry, it's still really useful.

Parameters:
    * prime.nvidiastring: String to use when nvidia is selected (defaults to 'intel')
    * prime.intelstring: String to use when intel is selected (defaults to 'intel')

Requires the following executable:
    * prime-select

contributed by `jeffeb3 <https://github.com/jeffeb3>`_ - many thanks!
"""

import core.module
import core.widget
import core.input

import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.query))

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.__chooseNvidia)
        core.input.register(self, button=core.input.RIGHT_MOUSE, cmd=self.__chooseIntel)

        self.nvidiastring = self.parameter("nvidiastring", "nv")
        self.intelstring = self.parameter("intelstring", "it")

    def __chooseNvidia(self, event):
        util.cli.execute("sudo prime-select nvidia")

    def __chooseIntel(self, event):
        util.cli.execute("sudo prime-select intel")

    def query(self, widget):
        try:
            res = util.cli.execute("prime-select query")
        except RuntimeError:
            return "n/a"

        for line in res.split("\n"):
            if not line:
                continue
            if "nvidia" in line:
                return self.nvidiastring
            if "intel" in line:
                return self.intelstring
        return "n/a"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays information about the current song in Google Play music player.

Requires the following executable:
    * gpmdp-remote

contributed by `TheEdgeOfRage <https://github.com/TheEdgeOfRage>`_ - many thanks!
"""

import core.module
import core.widget
import core.input

import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        widgets = [
            core.widget.Widget(name="gpmdp.prev"),
            core.widget.Widget(name="gpmdp.main", full_text=self.description),
            core.widget.Widget(name="gpmdp.next"),
        ]
        super().__init__(config, theme, widgets)

        core.input.register(
            widgets[0], button=core.input.LEFT_MOUSE, cmd="playerctl previous"
        )
        core.input.register(
            widgets[1], button=core.input.LEFT_MOUSE, cmd="playerctl play-pause"
        )
        core.input.register(
            widgets[2], button=core.input.LEFT_MOUSE, cmd="playerctl next"
        )

        self.__status = None
        self.__tags = None

    def description(self, widget):
        return self.__tags if self.__tags else "n/a"

    def update(self):
        self.__load_song()

    def state(self, widget):
        if widget.name == "gpmdp.prev":
            return "prev"
        if widget.name == "gpmdp.next":
            return "next"
        return self.__status

    def __load_song(self):
        info = util.cli.execute("gpmdp-remote current", ignore_errors=True)
        status = util.cli.execute("gpmdp-remote status", ignore_errors=True)
        self.__status = status.split("\n")[0].lower()
        self.__tags = info.split("\n")[0]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Displays network traffic
   * No extra configuration needed

contributed by `izn <https://github.com/izn>`_ - many thanks!
"""

import psutil
import netifaces

import core.module
import core.widget

import util.format

WIDGET_NAME = "network_traffic"


class Module(core.module.Module):
    def __init__(self, config, theme):
        widgets = [
            core.widget.Widget(
                name="{0}.rx".format(WIDGET_NAME),
                full_text=self.download_rate,
            ),
            core.widget.Widget(
                name="{0}.tx".format(WIDGET_NAME),
                full_text=self.upload_rate,
            ),
        ]
        super().__init__(config, theme, widgets)

        self.widgets()[0].set("theme.minwidth", "0000000KiB/s")
        self.widgets()[1].set("theme.minwidth", "0000000KiB/s")

        try:
            self._bandwidth = BandwidthInfo()

            self._rate_recv = "?"
            self._rate_sent = "?"
            self._bytes_recv = self._bandwidth.bytes_recv()
            self._bytes_sent = self._bandwidth.bytes_sent()
        except Exception:
            """ We do not want do explode anything """
            pass

    def state(self, widget):
        """Return the widget state"""

        if widget.name == "{}.rx".format(WIDGET_NAME):
            return "rx"
        elif widget.name == "{}.tx".format(WIDGET_NAME):
            return "tx"

        return None

    def update(self):
        try:
            bytes_recv = self._bandwidth.bytes_recv()
            bytes_sent = self._bandwidth.bytes_sent()

            self._rate_recv = bytes_recv - self._bytes_recv
            self._rate_sent = bytes_sent - self._bytes_sent

            self._bytes_recv, self._bytes_sent = bytes_recv, bytes_sent
        except Exception:
            """ We do not want do explode anything """
            pass

    def download_rate(self, _):
        return "{}/s".format(util.format.byte(self._rate_recv))

    def upload_rate(self, _):
        return "{}/s".format(util.format.byte(self._rate_sent))


class BandwidthInfo(object):
    """Get received/sent bytes from network adapter"""

    def bytes_recv(self):
        """Return received bytes"""
        return self.bandwidth().bytes_recv

    def bytes_sent(self):
        """Return sent bytes"""
        return self.bandwidth().bytes_sent

    def bandwidth(self):
        """Return bandwidth information"""
        io_counters = self.io_counters()
        return io_counters[self.default_network_adapter()]

    @classmethod
    def default_network_adapter(cls):
        """Return default active network adapter"""
        gateway = netifaces.gateways()["default"]

        if not gateway:
            raise "No default gateway found"

        return gateway[netifaces.AF_INET][1]

    @classmethod
    def io_counters(cls):
        """Return IO counters"""
        return psutil.net_io_counters(pernic=True)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays APT package update information (<to upgrade>/<to remove >)
Requires the following packages:

    * aptitude

contributed by `qba10 <https://github.com/qba10>`_ - many thanks!
"""

import re
import threading

import core.module
import core.widget
import core.decorators

import util.cli

PATTERN = "{} packages upgraded, {} newly installed, {} to remove and {} not upgraded."


def parse_result(to_parse):
    # We want the line with the package upgrade information
    line_to_parse = to_parse.split("\n")[-4]
    result = re.search(
        r"(.+) packages upgraded, (.+) newly installed, (.+) to remove", line_to_parse
    )

    return int(result.group(1)), int(result.group(3))


def get_apt_check_info(module):
    widget = module.widget()
    try:
        res = util.cli.execute("aptitude full-upgrade --simulate --assume-yes")
        widget.set("error", None)
    except (RuntimeError, FileNotFoundError) as e:
        widget.set("error", "unable to query APT: {}".format(e))
        return

    to_upgrade = 0
    to_remove = 0
    try:
        to_upgrade, to_remove = parse_result(res)
        widget.set("to_upgrade", to_upgrade)
        widget.set("to_remove", to_remove)
    except Exception as e:
        widget.set("error", "parse error: {}".format(e))

    core.event.trigger("update", [module.id], redraw_only=True)


class Module(core.module.Module):
    @core.decorators.every(minutes=30)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.updates))
        self.__thread = None

    def updates(self, widget):
        if widget.get("error"):
            return widget.get("error")
        return "{} to upgrade, {} to remove".format(
            widget.get("to_upgrade", 0), widget.get("to_remove", 0)
        )

    def update(self):
        if self.__thread and self.__thread.isAlive():
            return

        self.__thread = threading.Thread(target=get_apt_check_info, args=(self,))
        self.__thread.start()

    def state(self, widget):
        cnt = 0
        ret = "good"
        for t in ["to_upgrade", "to_remove"]:
            cnt += widget.get(t, 0)
        if cnt > 50:
            ret = "critical"
        elif cnt > 0:
            ret = "warning"
        if widget.get("error"):
            ret = "critical"

        return ret


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



from .pulseaudio import Module


class Module(Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, "source")


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the number of todo items from a text file

Parameters:
    * todo.file: File to read TODOs from (defaults to ~/Documents/todo.txt)


contributed by `codingo <https://github.com/codingo>`_ - many thanks!
"""

import os.path

import core.module
import core.widget
import core.input


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.output))

        self.__doc = os.path.expanduser(self.parameter("file", "~/Documents/todo.txt"))
        self.__todos = self.count_items()
        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd="xdg-open {}".format(self.__doc)
        )

    def output(self, widget):
        return str(self.__todos)

    def update(self):
        self.__todos = self.count_items()

    def state(self, widgets):
        if self.__todos == 0:
            return "empty"
        return "items"

    def count_items(self):
        try:
            i = -1
            with open(self.__doc) as f:
                for i, l in enumerate(f):
                    pass
            return i + 1
        except Exception:
            return 0


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays system load.

Parameters:
    * load.warning : Warning threshold for the one-minute load average (defaults to 70% of the number of CPUs)
    * load.critical: Critical threshold for the one-minute load average (defaults to 80% of the number of CPUs)
"""

import os
import multiprocessing

import core.module
import core.input


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.load))
        self._load = [0, 0, 0]
        try:
            self._cpus = multiprocessing.cpu_count()
        except NotImplementedError as e:
            self._cpus = 1
        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd="gnome-system-monitor"
        )

    def load(self, widget):
        return "{:.02f}/{:.02f}/{:.02f}".format(
            self._load[0], self._load[1], self._load[2]
        )

    def update(self):
        self._load = os.getloadavg()

    def state(self, widget):
        return self.threshold_state(self._load[0], self._cpus * 0.7, self._cpus * 0.8)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import os
import sys


def discover():
    libdir = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "bumblebee_status")
    )
    sys.path.append(libdir)


def utility(name):
    current_path = os.path.dirname(os.path.abspath(__file__))

    for path in [
        os.path.join(current_path, "..", "bin"),
        os.path.join(
            current_path, "..", "..", "..", "..", "share", "bumblebee-status", "utility"
        ),
        "/usr/share/bumblebee-status/bin/",
    ]:
        if os.path.exists(os.path.abspath(os.path.join(path, name))):
            return os.path.abspath(os.path.join(path, name))
    raise Exception("{} not found".format(name))


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Toggle dunst notifications.

contributed by `eknoes <https://github.com/eknoes>`_ - many thanks!
"""

import core.module
import core.widget
import core.input

import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(""))
        self._paused = False
        # Make sure that dunst is currently not paused
        util.cli.execute("killall -s SIGUSR2 dunst", ignore_errors=True)
        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.toggle_status)

    def toggle_status(self, event):
        self._paused = not self._paused

        try:
            if self._paused:
                util.cli.execute("killall -s SIGUSR1 dunst")
            else:
                util.cli.execute("killall -s SIGUSR2 dunst")
        except:
            self._paused = not self._paused  # toggling failed

    def state(self, widget):
        if self._paused:
            return ["muted", "warning"]
        return ["unmuted"]



import uuid
import logging

import core.event

import util.cli

LEFT_MOUSE = 1
MIDDLE_MOUSE = 2
RIGHT_MOUSE = 3
WHEEL_UP = 4
WHEEL_DOWN = 5


def button_name(button):
    if button == LEFT_MOUSE:
        return "left-mouse"
    if button == RIGHT_MOUSE:
        return "right-mouse"
    if button == MIDDLE_MOUSE:
        return "middle-mouse"
    if button == WHEEL_UP:
        return "wheel-up"
    if button == WHEEL_DOWN:
        return "wheel-down"
    return "n/a"


class Object(object):
    def __init__(self):
        super(Object, self).__init__()
        self.id = str(uuid.uuid4())


def __event_id(obj_id, button):
    return "{}::{}".format(obj_id, button_name(button))


def __execute(event, cmd, wait=False):
    try:
        util.cli.execute(
            cmd.format(instance=event.get("instance", ""), name=event.get("name", ""),),
            wait=wait,
            shell=True,
        )
    except Exception as e:
        logging.error("failed to invoke callback: {}".format(e))


def register(obj, button=None, cmd=None, wait=False):
    event_id = __event_id(obj.id if obj is not None else "", button)
    logging.debug("registering callback {}".format(event_id))
    if callable(cmd):
        core.event.register(event_id, cmd)
    else:
        core.event.register(event_id, lambda event: __execute(event, cmd, wait))


def trigger(event):
    if not "button" in event:
        return

    triggered = False
    for field in ["instance", "name"]:
        if not field in event:
            continue
        if core.event.trigger(__event_id(event[field], event["button"]), event):
            triggered = True
    if not triggered:
        core.event.trigger(__event_id("", event["button"]), event)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Enables handy interaction with arandr for display management.  Left-clicking
will execute arandr for interactive display management.  Right-clicking will
bring up a context- and state-sensitive menu that will allow you to switch to a
saved screen layout as well as toggle on/off individual connected displays.

Parameters:
    * No configuration parameters

Requires the following executable:
    * arandr
    * xrandr

contributed by `zerorust <https://github.com/zerorust>`_ - many thanks!
"""

import fnmatch
from functools import partial
import logging
import os
import re

import core.module
import core.widget
import core.input
import core.decorators
from util import popup
from util.cli import execute


log = logging.getLogger(__name__)

__screenlayout_dir__ = os.path.expanduser("~/.screenlayout")


class Module(core.module.Module):
    @core.decorators.never
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(''))

        self.manager = self.parameter("manager", "arandr")
        self.toggle_cmd = "xrandr"
        core.input.register(
            self,
            button=core.input.LEFT_MOUSE,
            cmd=self.popup,
        )

        core.input.register(self, button=core.input.RIGHT_MOUSE,
                            cmd=self.popup)

    @staticmethod
    def activate_layout(layout_path):
        log.debug("activating layout")
        log.debug(layout_path)
        execute(layout_path)

    def popup(self, widget):
        """Create Popup that allows the user to control their displays in one
        of three ways: launch arandr, select a pre-set screenlayout, toggle a
        display.
        """
        menu = popup.menu()
        menu.add_menuitem(
            "arandr",
            callback=partial(execute, self.manager)
        )
        menu.add_separator()

        displays = Module._get_displays()
        log.debug(displays)
        layouts = Module._get_layouts()
        available_layouts = Module._prune_layouts(layouts, displays)
        log.debug("Available layouts:")
        log.debug(available_layouts)

        if len(available_layouts) > 0:
            for layout in available_layouts:
                sh = os.path.join(__screenlayout_dir__, layout)
                sh_name = os.path.splitext(layout)[0]
                menu.add_menuitem(sh_name,
                                  callback=partial(self.activate_layout, sh))

        menu.add_separator()
        count_on = 0
        for display, state in displays.items():
            if state[1]:
                count_on += 1
        for display, state in displays.items():
            if not state[0]:
                continue
            on_off = "On" if state[1] else "Off"
            menu_line = "{}: {}".format(display, on_off)
            menu.add_menuitem(menu_line,
                              callback=partial(self.toggle_display, display,
                                               state[1], count_on))

        menu.show(widget, 0, 0)

    def toggle_display(self, display, current_state, count_on):
        """Toggle a display on or off based on its current state."""
        if current_state:
            log.debug("toggling off {}".format(display))
            if count_on == 1:
                log.info("attempted to turn off last display")
                return
            execute("{} --output {} --off".format(self.toggle_cmd, display))
        else:
            log.debug("toggling on {}".format(display))
            execute(
                "{} --output {} --auto".format(self.toggle_cmd, display)
            )

    @staticmethod
    def _get_displays():
        """Queries xrandr and builds a dict of the displays and their state.

        The dict entries are key by the display and are bools (True if
        connected).
        """
        displays = {}
        for line in execute("xrandr -q").split("\n"):
            if "connected" not in line:
                continue
            is_on = bool(re.search(r"\d+x\d+\+(\d+)\+\d+", line))
            parts = line.split(" ", 2)
            display = parts[0]
            displays[display] = (
                (True, is_on) if parts[1] == "connected" else (False, is_on)
            )

        return displays

    @staticmethod
    def _get_layouts():
        """Loads and parses the arandr screen layout scripts."""
        layouts = {}
        for filename in os.listdir(__screenlayout_dir__):
            if fnmatch.fnmatch(filename, '*.sh'):
                fullpath = os.path.join(__screenlayout_dir__, filename)
                with open(fullpath, "r") as file:
                    for line in file:
                        s_line = line.strip()
                        if "xrandr" not in s_line:
                            continue
                        displays_in_file = Module._parse_layout(line)
                        layouts[filename] = displays_in_file
        return layouts

    @staticmethod
    def _parse_layout(line):
        """Parses a single xrandr line to find what displays are active in the
        command.  Returns them as a list.
        """
        active_displays = []
        to_check = line[7:].split("--output ")
        for check in to_check:
            if not check or "off" in check:
                continue
            active_displays.append(check.split(" ")[0])
        return active_displays

    @staticmethod
    def _prune_layouts(layouts, displays):
        """Return a list of layouts whose displays are actually connected."""
        available = []
        for layout, needs in layouts.items():
            still_valid = True
            for need in needs:
                if need not in displays or not displays[need][0]:
                    still_valid = False
                    break
            if still_valid:
                available.append(layout)
        return available


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import sys
import json
import time

import core.theme
import core.event

import util.format


def dump_json(obj):
    return obj.dict()


def assign(src, dst, key, src_key=None, default=None):
    if not src_key:
        if key.startswith("_"):
            src_key = key
        else:
            src_key = key.replace("_", "-")  # automagically replace _ with -

    for k in src_key if isinstance(src_key, list) else [src_key]:
        if k in src:
            dst[key] = src[k]
            return
    if default is not None:
        dst[key] = default


class block(object):
    __COMMON_THEME_FIELDS = [
        "separator",
        "separator-block-width",
        "default-separators",
        "border-top",
        "border-left",
        "border-right",
        "border-bottom",
        "fg",
        "bg",
        "padding",
        "prefix",
        "suffix",
    ]

    def __init__(self, theme, module, widget):
        self.__attributes = {}
        for key in self.__COMMON_THEME_FIELDS:
            tmp = theme.get(key, widget)
            if tmp is not None:
                self.__attributes[key] = tmp

        self.__attributes["name"] = module.id
        self.__attributes["instance"] = widget.id
        self.__attributes["prev-bg"] = theme.get("bg", "previous")

    def set(self, key, value):
        self.__attributes[key] = value

    def is_pango(self, attr):
        if isinstance(attr, dict) and "pango" in attr:
            return True
        return False

    def pangoize(self, text):
        if not self.is_pango(text):
            return text
        self.__attributes["markup"] = "pango"
        attr = dict(text["pango"])
        text = attr.get("full_text", "")
        if "full_text" in attr:
            del attr["full_text"]
        result = "<span"
        for key, value in attr.items():
            result = '{} {}="{}"'.format(result, key, value)
        result = "{}>{}</span>".format(result, text)
        return result

    def dict(self):
        result = {}

        assign(self.__attributes, result, "full_text", ["full_text", "separator"])
        assign(self.__attributes, result, "separator", "default-separators")

        if "_decorator" in self.__attributes:
            assign(self.__attributes, result, "color", "bg")
            assign(self.__attributes, result, "background", "prev-bg")
            result["_decorator"] = True
        else:
            assign(self.__attributes, result, "color", "fg")
            assign(self.__attributes, result, "background", "bg")

        if "full_text" in self.__attributes:
            result["full_text"] = self.pangoize(result["full_text"])
            result["full_text"] = self.__format(self.__attributes["full_text"])

        for k in [
            "name",
            "instance",
            "separator_block_width",
            "border",
            "border_top",
            "border_bottom",
            "border_left",
            "border_right",
            "markup",
            "_raw",
            "_suffix",
            "_prefix",
            "min_width",
            "align",
        ]:
            assign(self.__attributes, result, k)

        return result

    def __pad(self, text):
        padding = self.__attributes.get("padding", "")
        if not text:
            return padding
        return "{}{}{}".format(padding, text, padding)

    def __format(self, text):
        if text is None:
            return None
        prefix = self.__pad(self.pangoize(self.__attributes.get("prefix")))
        suffix = self.__pad(self.pangoize(self.__attributes.get("suffix")))
        self.set("_prefix", prefix)
        self.set("_suffix", suffix)
        self.set("_raw", text)
        return "{}{}{}".format(prefix, text, suffix)


class i3(object):
    def __init__(self, theme=core.theme.Theme(), config=core.config.Config([])):
        self.__modules = []
        self.__content = {}
        self.__theme = theme
        self.__config = config
        core.event.register("update", self.update)
        core.event.register("start", self.draw, "start")
        core.event.register("draw", self.draw, "statusline")
        core.event.register("stop", self.draw, "stop")

    def content(self):
        return self.__content

    def theme(self, new_theme=None):
        if new_theme:
            self.__theme = new_theme
        return self.__theme

    def modules(self, modules=None):
        if not modules:
            return self.__modules
        self.__modules = modules if isinstance(modules, list) else [modules]

    def toggle_minimize(self, event):
        widget_id = event["instance"]

        if widget_id in self.__content:
            self.__content[widget_id]["minimized"] = not self.__content[widget_id]["minimized"]

    def draw(self, what, args=None):
        cb = getattr(self, what)
        data = cb(args) if args else cb()
        if "blocks" in data:
            sys.stdout.write(json.dumps(data["blocks"], default=dump_json))
        if "suffix" in data:
            sys.stdout.write(data["suffix"])
        sys.stdout.write("\n")
        sys.stdout.flush()

    def start(self):
        return {
            "blocks": {"version": 1, "click_events": True},
            "suffix": "\n[",
        }

    def stop(self):
        return {"suffix": "\n]"}

    def separator_block(self, module, widget):
        if not self.__theme.get("separator"):
            return []
        blk = block(self.__theme, module, widget)
        blk.set("_decorator", True)
        return [blk]

    def __content_block(self, module, widget):
        blk = block(self.__theme, module, widget)
        minwidth = widget.theme("minwidth")
        if minwidth is not None:
            try:
                blk.set("min-width", "-" * int(minwidth))
            except:
                blk.set("min-width", minwidth)
        blk.set("align", widget.theme("align"))
        blk.set("full_text", "\u2026" if self.__content[widget.id]["minimized"] else self.__content[widget.id]["text"])
        if widget.get("pango", False):
            blk.set("markup", "pango")
        if self.__config.debug():
            state = module.state(widget)
            if isinstance(state, list):
                state = ", ".join(state)
            blk.set("__state", state)
        return blk

    def blocks(self, module):
        blocks = []
        for widget in module.widgets():
            if widget.module and self.__config.autohide(widget.module.name):
                if not any(
                    state in widget.state() for state in ["warning", "critical"]
                ):
                    continue
            if module.hidden():
                continue
            blocks.extend(self.separator_block(module, widget))
            blocks.append(self.__content_block(module, widget))
            core.event.trigger("next-widget")
        return blocks

    # TODO: only updates full text, not the state!?
    def update(self, affected_modules=None, redraw_only=False):
        now = time.time()
        for module in self.__modules:
            if affected_modules and not module.id in affected_modules:
                continue
            if not affected_modules and module.next_update:
                if now < module.next_update:
                    continue

            if not redraw_only:
                module.update_wrapper()
                if module.parameter("interval", "") != "never":
                    module.next_update = now + util.format.seconds(
                        module.parameter("interval", self.__config.interval())
                    )
                else:
                    module.next_update = sys.maxsize
            for widget in module.widgets():
                if not widget.id in self.__content:
                    self.__content[widget.id] = { "minimized": False }
                self.__content[widget.id]["text"] = widget.full_text()

    def statusline(self):
        blocks = []
        for module in self.__modules:
            blocks.extend(self.blocks(module))
        return {"blocks": blocks, "suffix": ","}

    def wait(self, interval):
        time.sleep(interval)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
